import openmdao.api as om
import numpy as np


class TimeODE(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # states
        self.add_input('sdot', val=np.zeros(nn), desc='distance along track', units='m/s')
        self.add_input('ndot', val=np.zeros(nn), desc='distance perpendicular to centerline',
                       units='m/s')
        self.add_input('alphadot', val=np.zeros(nn), desc='angle relative to centerline',
                       units='rad/s')
        self.add_input('Vdot', val=np.zeros(nn), desc='speed', units='m/s**2')
        self.add_input('lambdadot', val=np.zeros(nn), desc='body slip angle', units='rad/s')
        self.add_input('omegadot', val=np.zeros(nn), desc='yaw rate', units='rad/s**2')
        self.add_input('axdot', val=np.zeros(nn), desc='longitudinal jerk', units='m/s**3')
        self.add_input('aydot', val=np.zeros(nn), desc='lateral jerk', units='m/s**3')

        # outputs
        self.add_output('dn_ds', val=np.zeros(nn), desc='distance perpendicular to centerline',
                        units='m/m')
        self.add_output('dalpha_ds', val=np.zeros(nn), desc='angle relative to centerline',
                        units='rad/m')
        self.add_output('dV_ds', val=np.zeros(nn), desc='speed', units='1/s')
        self.add_output('dlambda_ds', val=np.zeros(nn), desc='body slip angle', units='rad/m')
        self.add_output('domega_ds', val=np.zeros(nn), desc='yaw rate', units='rad/(s*m)')
        self.add_output('dax_ds', val=np.zeros(nn), desc='longitudinal jerk', units='1/s**2')
        self.add_output('day_ds', val=np.zeros(nn), desc='lateral jerk', units='1/s**2')

        # Setup partials
        arange = np.arange(self.options['num_nodes'], dtype=int)

        # partials

        self.declare_partials(of='dn_ds', wrt='ndot', rows=arange, cols=arange)
        self.declare_partials(of='dn_ds', wrt='sdot', rows=arange, cols=arange)

        self.declare_partials(of='dalpha_ds', wrt='alphadot', rows=arange, cols=arange)
        self.declare_partials(of='dalpha_ds', wrt='sdot', rows=arange, cols=arange)

        self.declare_partials(of='dV_ds', wrt='Vdot', rows=arange, cols=arange)
        self.declare_partials(of='dV_ds', wrt='sdot', rows=arange, cols=arange)

        self.declare_partials(of='dlambda_ds', wrt='lambdadot', rows=arange, cols=arange)
        self.declare_partials(of='dlambda_ds', wrt='sdot', rows=arange, cols=arange)

        self.declare_partials(of='domega_ds', wrt='omegadot', rows=arange, cols=arange)
        self.declare_partials(of='domega_ds', wrt='sdot', rows=arange, cols=arange)

        self.declare_partials(of='dax_ds', wrt='axdot', rows=arange, cols=arange)
        self.declare_partials(of='dax_ds', wrt='sdot', rows=arange, cols=arange)

        self.declare_partials(of='day_ds', wrt='aydot', rows=arange, cols=arange)
        self.declare_partials(of='day_ds', wrt='sdot', rows=arange, cols=arange)

    def compute(self, inputs, outputs):
        omegadot = inputs['omegadot']
        sdot = inputs['sdot']
        Vdot = inputs['Vdot']
        ndot = inputs['ndot']
        lambdadot = inputs['lambdadot']
        alphadot = inputs['alphadot']
        axdot = inputs['axdot']
        aydot = inputs['aydot']

        outputs['domega_ds'] = omegadot/sdot
        outputs['dV_ds'] = Vdot/sdot
        outputs['dalpha_ds'] = alphadot/sdot
        outputs['dlambda_ds'] = lambdadot/sdot
        outputs['dn_ds'] = ndot/sdot
        outputs['dax_ds'] = axdot/sdot
        outputs['day_ds'] = aydot/sdot

    def compute_partials(self, inputs, jacobian):
        omegadot = inputs['omegadot']
        sdot = inputs['sdot']
        Vdot = inputs['Vdot']
        ndot = inputs['ndot']
        lambdadot = inputs['lambdadot']
        alphadot = inputs['alphadot']
        axdot = inputs['axdot']
        aydot = inputs['aydot']

        jacobian['dn_ds', 'sdot'] = -ndot/sdot**2
        jacobian['dn_ds', 'ndot'] = 1/sdot

        jacobian['dalpha_ds', 'sdot'] = -alphadot/sdot**2
        jacobian['dalpha_ds', 'alphadot'] = 1/sdot

        jacobian['domega_ds', 'sdot'] = -omegadot/sdot**2
        jacobian['domega_ds', 'omegadot'] = 1/sdot

        jacobian['dlambda_ds', 'sdot'] = -lambdadot/sdot**2
        jacobian['dlambda_ds', 'lambdadot'] = 1/sdot

        jacobian['dV_ds', 'sdot'] = -Vdot/sdot**2
        jacobian['dV_ds', 'Vdot'] = 1/sdot

        jacobian['dax_ds', 'sdot'] = -axdot/sdot**2
        jacobian['dax_ds', 'axdot'] = 1/sdot

        jacobian['day_ds', 'sdot'] = -aydot/sdot**2
        jacobian['day_ds', 'aydot'] = 1/sdot



import unittest

import numpy as np

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse

import dymos as dm
from dymos.utils.lgl import lgl
from dymos.models.eom import FlightPathEOM2D


@use_tempdirs
class TestParameterConnections(unittest.TestCase):

    @require_pyoptsparse(optimizer='SLSQP')
    def test_dynamic_parameter_connections_radau(self):

        class TrajectoryODE(om.Group):
            def initialize(self):
                self.options.declare('num_nodes', types=int)

            def setup(self):
                nn = self.options['num_nodes']

                self.add_subsystem('sum', om.ExecComp('m_tot = sum(m)',
                                                      m={'val': np.zeros((nn, 2, 2)),
                                                         'units': 'kg'},
                                                      m_tot={'val': np.zeros(nn),
                                                             'units': 'kg'}))

                self.add_subsystem('eom', FlightPathEOM2D(num_nodes=nn))

                self.connect('sum.m_tot', 'eom.m')

        optimizer = 'SLSQP'
        num_segments = 1
        transcription_order = 5

        p = om.Problem(model=om.Group())

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = optimizer
        p.driver.declare_coloring()

        seg_ends, _ = lgl(num_segments + 1)

        phase = dm.Phase(ode_class=TrajectoryODE,
                         transcription=dm.Radau(num_segments=num_segments, order=transcription_order,
                                                segment_ends=seg_ends))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(initial_bounds=(0.0, 100.0), duration_bounds=(0., 100.), units='s')

        phase.add_state('h', fix_initial=True, fix_final=True, lower=0.0, units='m', rate_source='eom.h_dot')
        phase.add_state('v', fix_initial=True, fix_final=False, units='m/s', rate_source='eom.v_dot')

        phase.add_parameter('m', val=[[1, 2], [3, 4]], units='kg', targets='sum.m')

        p.model.linear_solver = om.DirectSolver()

        p.setup(check=True, force_alloc_complex=True)

        phase.set_time_val(initial=0.0, duration=100.0)
        phase.set_state_val('h', [20, 0])
        phase.set_state_val('v', [0, -5])

        p.run_model()

        expected = np.broadcast_to(np.array([[1, 2], [3, 4]]),
                                   (p.model.phase0.options['transcription'].grid_data.num_nodes, 2, 2))
        assert_near_equal(p.get_val('phase0.rhs_all.sum.m'), expected)

    @require_pyoptsparse(optimizer='SLSQP')
    def test_static_parameter_connections_radau(self):

        class TrajectoryODE(om.Group):
            def initialize(self):
                self.options.declare('num_nodes', types=int)

            def setup(self):
                nn = self.options['num_nodes']

                self.add_subsystem('sum', om.ExecComp('m_tot = sum(m)',
                                                      m={'val': np.zeros((2, 2)),
                                                         'units': 'kg'},
                                                      m_tot={'val': np.zeros(nn),
                                                             'units': 'kg'}))

                self.add_subsystem('eom', FlightPathEOM2D(num_nodes=nn))

                self.connect('sum.m_tot', 'eom.m')

        optimizer = 'SLSQP'
        num_segments = 1
        transcription_order = 5

        p = om.Problem(model=om.Group())

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = optimizer
        p.driver.declare_coloring()

        seg_ends, _ = lgl(num_segments + 1)

        phase = dm.Phase(ode_class=TrajectoryODE,
                         transcription=dm.Radau(num_segments=num_segments,
                                                order=transcription_order,
                                                segment_ends=seg_ends))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(initial_bounds=(0.0, 100.0), duration_bounds=(0., 100.))

        phase.add_state('h', fix_initial=True, fix_final=True, lower=0.0, units='m', rate_source='eom.h_dot')
        phase.add_state('v', fix_initial=True, fix_final=False, units='m/s', rate_source='eom.v_dot')

        phase.add_parameter('m', val=[[1, 2], [3, 4]], units='kg', targets='sum.m', static_target=True)

        p.model.linear_solver = om.DirectSolver()

        p.setup(check=True, force_alloc_complex=True)

        phase.set_time_val(initial=0.0, duration=100.0)
        phase.set_state_val('h', [20, 0])
        phase.set_state_val('v', [0, -5])

        p.run_model()

        expected = np.array([[1, 2], [3, 4]])
        assert_near_equal(p.get_val('phase0.rhs_all.sum.m'), expected)

    @require_pyoptsparse(optimizer='SLSQP')
    def test_dynamic_parameter_connections_gl(self):

        class TrajectoryODE(om.Group):
            def initialize(self):
                self.options.declare('num_nodes', types=int)

            def setup(self):
                nn = self.options['num_nodes']

                self.add_subsystem('sum', om.ExecComp('m_tot = sum(m)',
                                                      m={'val': np.zeros((nn, 2, 2)),
                                                         'units': 'kg'},
                                                      m_tot={'val': np.zeros(nn),
                                                             'units': 'kg'}))

                self.add_subsystem('eom', FlightPathEOM2D(num_nodes=nn))

                self.connect('sum.m_tot', 'eom.m')

        optimizer = 'SLSQP'
        num_segments = 1
        transcription_order = 5

        p = om.Problem(model=om.Group())

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = optimizer
        p.driver.declare_coloring()

        seg_ends, _ = lgl(num_segments + 1)

        phase = dm.Phase(ode_class=TrajectoryODE,
                         transcription=dm.GaussLobatto(num_segments=num_segments,
                                                       order=transcription_order,
                                                       segment_ends=seg_ends))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(initial_bounds=(0.0, 100.0), duration_bounds=(0., 100.), units='s')

        phase.add_state('h', fix_initial=True, fix_final=True, lower=0.0, units='m', rate_source='eom.h_dot')
        phase.add_state('v', fix_initial=True, fix_final=False, units='m/s', rate_source='eom.v_dot')

        phase.add_parameter('m', val=[[1, 2], [3, 4]], units='kg', targets='sum.m')

        p.model.linear_solver = om.DirectSolver()

        p.setup(check=True, force_alloc_complex=True)

        phase.set_time_val(initial=0.0, duration=100.0)
        phase.set_state_val('h', [20, 0])
        phase.set_state_val('v', [0, -5])

        p.run_model()

        gd = p.model.phase0.options['transcription'].grid_data

        expected = np.broadcast_to(np.array([[1, 2], [3, 4]]),
                                   (gd.subset_num_nodes['state_disc'], 2, 2))
        assert_near_equal(p.get_val('phase0.rhs_disc.sum.m'), expected)

        expected = np.broadcast_to(np.array([[1, 2], [3, 4]]),
                                   (gd.subset_num_nodes['col'], 2, 2))
        assert_near_equal(p.get_val('phase0.rhs_col.sum.m'), expected)

    @require_pyoptsparse(optimizer='SLSQP')
    def test_static_parameter_connections_gl(self):

        class TrajectoryODE(om.Group):
            def initialize(self):
                self.options.declare('num_nodes', types=int)

            def setup(self):
                nn = self.options['num_nodes']

                self.add_subsystem('sum', om.ExecComp('m_tot = sum(m)',
                                                      m={'val': np.zeros((2, 2)),
                                                         'units': 'kg'},
                                                      m_tot={'val': np.zeros(nn),
                                                             'units': 'kg'}))

                self.add_subsystem('eom', FlightPathEOM2D(num_nodes=nn))

                self.connect('sum.m_tot', 'eom.m')

        optimizer = 'SLSQP'
        num_segments = 1
        transcription_order = 5

        p = om.Problem(model=om.Group())

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = optimizer
        p.driver.declare_coloring()

        seg_ends, _ = lgl(num_segments + 1)

        phase = dm.Phase(ode_class=TrajectoryODE,
                         transcription=dm.GaussLobatto(num_segments=num_segments,
                                                       order=transcription_order,
                                                       segment_ends=seg_ends))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(initial_bounds=(0.0, 100.0), duration_bounds=(0., 100.), units='s')

        phase.add_state('h', fix_initial=True, fix_final=True, lower=0.0, units='m', rate_source='eom.h_dot')
        phase.add_state('v', fix_initial=True, fix_final=False, units='m/s', rate_source='eom.v_dot')

        phase.add_parameter('m', val=[[1, 2], [3, 4]], units='kg', targets='sum.m', static_target=True)

        p.model.linear_solver = om.DirectSolver()

        p.setup(check=True, force_alloc_complex=True)

        phase.set_time_val(initial=0.0, duration=100.0)
        phase.set_state_val('h', [20, 0])
        phase.set_state_val('v', [0, -5])

        p.run_model()

        expected = np.array([[1, 2], [3, 4]])
        assert_near_equal(p.get_val('phase0.rhs_disc.sum.m'), expected)
        assert_near_equal(p.get_val('phase0.rhs_col.sum.m'), expected)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import numpy as np


def animate_cartpole(x, theta, force, interval=20, force_scaler=0.1, save_gif=False, gif_fps=20):

    import matplotlib.pyplot as plt
    import matplotlib.patches as patches
    from matplotlib import animation
    # x: time history of cart location, 1d vector
    # theta: time history of pole angle, 1d vector
    # force: control input force

    # keep showing the final state for 100 ms
    extend = 500 // interval
    x = np.concatenate((x, np.ones(extend) * x[-1]))
    theta = np.concatenate((theta, np.ones(extend) * theta[-1]))
    force = np.concatenate((force, np.ones(extend) * force[-1]))

    # cart parameters
    cart_width = 0.25
    cart_height = 0.15
    lpole = 0.5

    # path of the pole tip
    x_pole = lpole * np.sin(theta) + x
    y_pole = -lpole * np.cos(theta) + cart_height / 2.0

    # force vector
    arrow_direc = force * force_scaler
    arrow_origin = x - arrow_direc
    arrow_y = cart_height / 2

    # x_lim and y_lim for figure
    xlim = [min(min(x), min(x_pole)) - cart_width / 2 - 0.1, max(max(x), max(x_pole)) + cart_width / 2 + 0.1]
    ylim = [cart_height / 2 - lpole - 0.05, cart_height / 2 + lpole + 0.05]

    fig, ax = plt.subplots()

    def init():
        # initialize figure
        ax.set_xlabel("x (m)")
        ax.set_ylabel("y (m)")
        ax.axis("equal")

    def animate(i):
        ax.clear()
        # plot "floor"
        ax.plot(xlim, [0.0, 0.0], color="k")
        ax.set_xlim(xlim)
        ax.set_ylim(ylim)
        # plot cart
        cart = ax.add_patch(
            patches.Rectangle((-cart_width / 2.0 + x[i], 0.0), cart_width, cart_height, edgecolor="C0", linewidth=1, fill=False)
        )
        # plot pole
        pole = ax.plot([x[i], x_pole[i]], [cart_height / 2, y_pole[i]], "o-", lw=2, color="C0")
        # plot pole tip path
        path = ax.plot(x_pole[:i], y_pole[:i], "--", lw=1, color="black")
        # plot force vector
        f = ax.arrow(
            arrow_origin[i],
            arrow_y,
            arrow_direc[i],
            0.0,
            color="C1",
            head_width=0.02,
            head_length=0.03,
            length_includes_head=True,
        )
        return pole, cart, path, f

    anim = animation.FuncAnimation(fig, animate, init_func=init, frames=len(x), interval=interval, repeat=True)
    if save_gif:
        anim.save("cartpole.gif", dpi=300, writer=animation.PillowWriter(fps=gif_fps))
    plt.show()



import io
from contextlib import redirect_stdout
import unittest

import numpy as np

import openmdao.api as om
from openmdao.utils.testing_utils import use_tempdirs
from openmdao.utils.assert_utils import assert_near_equal

import dymos as dm
from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE


@use_tempdirs
class TestDuplicateTimeseriesInput(unittest.TestCase):

    def _make_problem(self, transcription):

        #
        # Define the OpenMDAO problem
        #
        p = om.Problem(model=om.Group())

        #
        # Define a Trajectory object
        #
        traj = dm.Trajectory()

        p.model.add_subsystem('traj', subsys=traj)

        #
        # Define a Dymos Phase object with GaussLobatto Transcription
        #
        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=transcription(num_segments=10, order=3))

        traj.add_phase(name='phase0', phase=phase)

        #
        # Set the time options
        # Time has no targets in our ODE.
        # We fix the initial time so that the it is not a design variable in the optimization.
        # The duration of the phase is allowed to be optimized, but is bounded on [0.5, 10].
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(1.0, 10.0), units='s')

        #
        # Set the time options
        # Initial values of positions and velocity are all fixed.
        # The final value of position are fixed, but the final velocity is a free variable.
        # The equations of motion are not functions of position, so 'x' and 'y' have no targets.
        # The rate source points to the output in the ODE which provides the time derivative of the
        # given state.
        phase.add_state('x', fix_initial=True, fix_final=True)
        phase.add_state('y', fix_initial=True, fix_final=True)
        phase.add_state('v', fix_initial=True)

        # Define theta as a control.
        phase.add_control(name='theta', units='rad', lower=0.01, upper=np.pi-0.01, shape=(1,))

        phase.add_timeseries_output('*')
        phase.add_timeseries_output('x', output_name='state_x')

        # Minimize final time.
        phase.add_objective('time', loc='final')

        # Set the driver.
        p.driver = om.ScipyOptimizeDriver()

        # Allow OpenMDAO to automatically determine our sparsity pattern.
        # Doing so can significant speed up the execution of Dymos.
        p.driver.declare_coloring()

        # Setup the problem
        p.setup(check=True)

        phase.set_time_val(initial=0.0, duration=2.0)

        phase.set_state_val('x', [0, 10], units='m')
        phase.set_state_val('y', [10, 5], units='m')
        phase.set_state_val('v', [0.0, 5], units='m/s')
        phase.set_control_val('theta', [5, 90], units='deg')

        return p

    def test_duplicate_timeseries_input_gl(self):
        p = self._make_problem(dm.GaussLobatto)

        with io.StringIO() as buf, redirect_stdout(buf):
            p.run_driver()
            output = buf.getvalue()

        # First check that the input duplication does not exist
        warning_str = 'WARNING: The following components have multiple inputs connected to ' \
                      'the same source, which can introduce unnecessary data transfer overhead:'

        if warning_str in output:
            raise AssertionError('Multiple inputs connected to the same source exist')

        # Now check that the outputs are still the same.
        nom_x = p.get_val('traj.phase0.timeseries.x').T
        added_x = p.get_val('traj.phase0.timeseries.state_x').T

        assert_near_equal(nom_x, added_x)

    def test_duplicate_timeseries_input_radau(self):
        p = self._make_problem(dm.Radau)

        with io.StringIO() as buf, redirect_stdout(buf):
            p.run_model()
            output = buf.getvalue()

        # First check that the input duplication does not exist
        warning_str = 'WARNING: The following components have multiple inputs connected to ' \
                      'the same source, which can introduce unnecessary data transfer overhead:'

        if warning_str in output:
            raise AssertionError('Multiple inputs connected to the same source exist')

        # Now check that the outputs are still the same.
        nom_x = p.get_val('traj.phase0.timeseries.x').T
        added_x = p.get_val('traj.phase0.timeseries.state_x').T

        assert_near_equal(nom_x, added_x)



import dymos as dm
from collections import OrderedDict
import inspect
import re
from pathlib import Path
from openmdao.visualization.htmlpp import HtmlPreprocessor
import openmdao.utils.reports_system as rptsys

CBN = 'children_by_name'
_default_linkage_report_title = 'Dymos Linkage Report'
_default_linkage_report_filename = 'linkage_report.html'


def _create_model_data(traj):
    """
    Creates the model_data dictionary for the given trajectory.

    Parameters
    ----------
    traj : A dymos Trajectory object.

    Returns
    -------
    dict
        The generated model_data.
    """
    model_data = _trajectory_to_dict(traj)
    model_data['connections_list'] = _linkages_to_list(traj, model_data)
    model_data['connections_list'].extend(_parameter_connections(traj, model_data))
    return model_data


def create_linkage_report(traj, output_file: str = _default_linkage_report_filename,
                          show_all_vars=True,
                          title=_default_linkage_report_title, embedded=False):
    """
    Create a tree based on the trajectory and linkages, then export as HTML.

    Parameters
    ----------
    traj : Trajectory
        The Dymos Trajectory containing the phase and linkage data.
    output_file : str
        Where to write the HTML file with the report diagram.
    show_all_vars : bool
        If True, display all parameters and controls even if not linked.
    title : str
        The title to display in the HTML document.
    embedded : bool
        Whether the file is to be included in another. If True, leave out some tags.
    """
    model_data = _create_model_data(traj)

    _convert_dicts_to_lists(model_data['tree'], show_all_vars)

    html_vars = {
        'title': title,
        'embeddable': "non-embedded-diagram",
        'sw_version': 0.1,
        'model_data': model_data
    }

    import openmdao
    openmdao_dir = Path(inspect.getfile(openmdao)).parent
    vis_dir = openmdao_dir / 'visualization' / 'n2_viewer'

    dymos_dir = Path(inspect.getfile(dm)).parent
    reports_dir = dymos_dir / 'visualization' / 'linkage'

    HtmlPreprocessor(str(reports_dir / 'report_template.html'), output_file,
                     search_path=[str(vis_dir), str(reports_dir)], allow_overwrite=True,
                     var_dict=html_vars, verbose=False).run()


def _is_fixed(var_name: str, class_name: str, phase, loc: str):
    """
    Determine whether a variable is fixed or not.

    Parameters
    ----------
    var_name : str
        Identifier of the variable as known to the phase.
    class_name : str
        The type of variable.
    phase : Phase
        The phase where the variable is found.
    loc : str
        Either 'initial' or 'final' for non-parameters.

    Returns
    -------
    bool
        True if the variable is fixed, otherwise False.
    """
    fixed = False

    if class_name == 'time':
        fixed = phase.is_time_fixed(loc)
    elif class_name == 'state':
        fixed = phase.is_state_fixed(var_name, loc)
    elif class_name == 'control':
        fixed = phase.is_control_fixed(var_name, loc)
    else:
        fixed = True

    return bool(fixed)


def _tree_var(var_name, phase, loc=None, var_name_prefix=""):
    """
    Create a dict to represent a variable in the tree.

    Parameters
    ----------
    var_name : str
        The name of the variable.
    phase : dymos.Phase
        Reference to the Phase object.
    loc : str
        Either 'initial' or 'final' for non-parameters.
    var_name_prefix : str
        A string to prepend to the variable name.

    Returns
    -------
    dict
        Information about the variable represented in the report.
    """
    class_name = str(phase.classify_var(var_name))
    opt = phase.parameter_options[var_name]['opt'] if class_name == 'parameter' else None

    return {
        'name': f"{var_name_prefix}{var_name}",
        'type': 'variable',
        'class': class_name,
        'fixed': _is_fixed(var_name, class_name, phase, loc),
        'paramOpt': opt
    }


def _trajectory_to_dict(traj):
    """
    Iterate over the variables in the trajectory's phases and create a hierarchical structure.

    Parameters
    ----------
    traj : Trajectory
        The trajectory containing the phases to find variables in.

    Returns
    -------
    dict
        A dictionary in the form: root -> phases -> conditions (initial/final) -> variables.
    """
    model_data = {
        'tree': {
            'name': 'Trajectory',
            'type': 'root',
            CBN: OrderedDict()
        }
    }

    tree = model_data['tree']

    tree[CBN]['params'] = {
        'name': 'params',
        'type': 'phase',
        CBN: OrderedDict()
    }

    for param_name, param in traj.parameter_options.items():
        tree[CBN]['params'][CBN][param_name] = {
            'name': str(param_name),
            'type': 'variable',
            'class': 'parameter',
            'fixed': False,
            'paramOpt': param['opt'],
            'connected': True
        }

    for phase_name, phase in traj._phases.items():
        tree_phase = {
            'name': str(phase_name),
            'type': 'phase',
            CBN: OrderedDict({
                'params': {'name': 'params', 'type': 'condition', CBN: OrderedDict()},
                'initial': {'name': 'initial', 'type': 'condition', CBN: OrderedDict()},
                'final': {'name': 'final', 'type': 'condition', CBN: OrderedDict()}
            })
        }

        tree[CBN][phase_name] = tree_phase

        params_children = tree_phase[CBN]['params'][CBN]
        condition_children = OrderedDict({
            'initial': tree_phase[CBN]['initial'][CBN],
            'final': tree_phase[CBN]['final'][CBN]
        })

        # Time
        for loc, child in condition_children.items():
            child['time'] = _tree_var('time', phase, loc)

        # States
        for state_name in phase.state_options:
            for loc, child in condition_children.items():
                child[state_name] = _tree_var(state_name, phase, loc, 'states:')

        # Controls
        for control_name in phase.control_options:
            for loc, child in condition_children.items():
                child[control_name] = _tree_var(control_name, phase, loc, 'controls:')

        # Parameters
        for param_name, param in phase.parameter_options.items():
            params_children[param_name] = _tree_var(param_name, phase)

    return model_data


def _linkages_to_list(traj, model_data):
    """
    Extract variables from the trajectory's linkages dict and put them in an array.

    Also flag in the model_data dict whether a variable is linked and/or fixed.

    Parameters
    ----------
    traj : Trajectory
        The trajectory with the linkages structure.
    model_data : dict
        Data extracted from the trajectory's phases.

    Returns
    -------
    list
        Dict objects with src, tgt, src_fixed, and tgt_fixed keys.
    """
    linkages = []
    tree = model_data['tree']

    for phase_pair, var_dict in traj._linkages.items():
        for var_pair, options in var_dict.items():
            skip = False
            loc = [options['loc_a'], options['loc_b']]
            tree_var = []

            for i in range(2):
                phase_cbn = tree[CBN][phase_pair[i]][CBN]
                if var_pair[i] not in phase_cbn[loc[i]][CBN]:
                    if var_pair[i] in phase_cbn['params'][CBN]:
                        # Variable is in the parameters area rather than initial or final condition
                        loc[i] = 'params'
                    else:
                        # When linking two ODE outputs, it's possible for the variable to not show up in the tree.
                        # For example, see the 'ke' linkage in
                        # TestTwoPhaseCannonballODEOutputLinkage.test_traj_param_target_unspecified_units.
                        # For now, let these linkages go undisplayed in the report.
                        skip = True

                if skip is False:
                    tree_var.append(phase_cbn[loc[i]][CBN][var_pair[i]])

            if skip is True:
                continue

            tree_var[0]['linked'] = tree_var[1]['linked'] = True
            tree_var[0]['connected'] = tree_var[1]['connected'] = options['connected']

            linkages.append({
                'src': f'{phase_pair[0]}.{loc[0]}.{tree_var[0]["name"]}',
                'src_fixed': tree_var[0]['fixed'],
                'tgt': f'{phase_pair[1]}.{loc[1]}.{tree_var[1]["name"]}',
                'tgt_fixed': tree_var[1]['fixed']
            })

    return linkages


def _conn_name_to_path(name):
    """ Convert the full name of the connection to a format the diagram uses. """
    if re.search(r'.*param_comp.parameter_vals.*|.*param_comp.parameters.*', name):
        # Special handling in case the name contains colons
        tokens = re.split(r'\.', name)
        last_name = tokens[-1]
        param_name = re.split(r':', last_name, 1)
        tokens[-1] = param_name[0]
        tokens.append(param_name[1])
    else:
        tokens = re.split(r'\W+', name)

    if len(tokens) > 0:
        if tokens[1] == 'param_comp':
            return f'params.{tokens.pop()}'

        # Example: traj.linkages.v1_to_vr:alpha -> v1_to_vr.alpha
        if tokens[1] == 'linkages':
            return f'{tokens[2]}.{tokens[3]}'

        if len(tokens) > 3 and tokens[3] == 'param_comp':
            return f'{tokens[2]}.params.{tokens.pop()}'

    return name


def _is_param_conn(name) -> bool:
    """ Determine if the specified connection involves a parameter. """
    return re.match(r'.*param_comp\.parameters:.*', name) \
        and not re.match(r'.*param_comp\.parameter_vals:.*', name)


def _is_ignored_conn(name) -> bool:
    """ Determine if an connection endpoint should be ignored for this diagram. """
    return re.match(r'.*timeseries.*', name) or re.match(r'.*_auto_ivc.*', name)


def _var_ref_from_path(tree, path: str):
    """ Find a reference into the tree from a path string. """
    tokens = re.split(r'\.', path)

    if (len(tokens) < 2):
        return None

    refpath = tree

    try:
        for t in tokens:
            refpath = refpath[CBN][t]
    except (KeyError):
        return None

    return refpath


def _get_fixed_val(tree, path: str) -> bool:
    """ Get the value of the fixed property for the specified path. """
    ref = _var_ref_from_path(tree, path)
    fixed = False if (ref is None or 'fixed' not in ref) else ref['fixed']

    return fixed


def _parameter_connections(traj, model_data):
    """ Find all parameter-to-parameter connections. """
    allconn = traj._problem_meta['model_ref']()._conn_global_abs_in2out
    tree = model_data['tree']
    param_conns = []

    for tgt, src in allconn.items():
        is_ignored = (_is_ignored_conn(src) or _is_ignored_conn(tgt))
        is_param_conn = (_is_param_conn(src) or _is_param_conn(tgt))
        if not is_ignored and is_param_conn:
            src_path = _conn_name_to_path(src)
            tgt_path = _conn_name_to_path(tgt)
            param_conns.append({
                'src': src_path,
                'src_fixed': _get_fixed_val(tree, src_path),
                'tgt': tgt_path,
                'tgt_fixed': _get_fixed_val(tree, tgt_path)
            })

    return param_conns


def _display_child(child, show_all_vars):
    """ Determine whether the object should be included in the diagram. """
    if show_all_vars is True:
        return True

    if CBN in child or 'children' in child:
        return True

    if 'linked' in child:
        return True

    if child['class'] == 'time' or child['class'] == 'state':
        return True

    return False


def _convert_dicts_to_lists(tree_dict, show_all_vars):
    """ Convert all children_by_name dicts to lists. """
    if CBN in tree_dict:
        tree_dict['children'] = []
        for child in tree_dict[CBN].values():
            _convert_dicts_to_lists(child, show_all_vars)
            if _display_child(child, show_all_vars):
                tree_dict['children'].append(child)

        tree_dict.pop(CBN)


def _run_linkage_report(prob):
    """ Function invoked by the reports system """

    # Find all Trajectory objects in the Problem. Usually, there's only one
    for traj in prob.model.system_iter(include_self=True, recurse=True, typ=dm.Trajectory):
        # Only create a report for a trajectory with linkages
        if traj._linkages:
            report_filename = f'{traj.pathname}_{_default_linkage_report_filename}'
            report_path = str(Path(prob.get_reports_dir()) / report_filename)
            create_linkage_report(traj, report_path)


def _linkage_report_register():
    rptsys.register_report('dymos.linkage', _run_linkage_report, _default_linkage_report_title,
                           'Problem', 'final_setup', 'post')
    rptsys._default_reports.append('dymos.linkage')



import numpy as np
import openmdao.api as om


class RobertsonODE(om.ExplicitComponent):
    """example for a stiff ODE from Robertson.
    """

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # input: state
        self.add_input('x', shape=nn, desc="state x", units=None)
        self.add_input('y', shape=nn, desc="state y", units=None)
        self.add_input('z', shape=nn, desc="state z", units=None)

        # output: derivative of state
        self.add_output('xdot', shape=nn, desc='derivative of x', units="1/s")
        self.add_output('ydot', shape=nn, desc='derivative of y', units="1/s")
        self.add_output('zdot', shape=nn, desc='derivative of z', units="1/s")

        r = np.arange(0, nn)
        self.declare_partials(of='*', wrt='*', method='exact', rows=r, cols=r)

    def compute(self, inputs, outputs):
        x = inputs['x']
        y = inputs['y']
        z = inputs['z']

        xdot = -0.04 * x + 1e4 * y * z
        zdot = 3e7 * y ** 2
        ydot = - (xdot + zdot)

        outputs['xdot'] = xdot
        outputs['ydot'] = ydot
        outputs['zdot'] = zdot

    def compute_partials(self, inputs, jacobian):
        # inputs['x']  # x is not needed to compute partials
        y = inputs['y']
        z = inputs['z']

        xdot_y = 1e4 * z
        xdot_z = 1e4 * y

        zdot_y = 6e7 * y

        jacobian['xdot', 'x'] = -0.04
        jacobian['xdot', 'y'] = xdot_y
        jacobian['xdot', 'z'] = xdot_z

        jacobian['ydot', 'x'] = 0.04
        jacobian['ydot', 'y'] = - (xdot_y + zdot_y)
        jacobian['ydot', 'z'] = - xdot_z

        jacobian['zdot', 'x'] = 0.0
        jacobian['zdot', 'y'] = zdot_y



import unittest

import numpy as np

from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs

import dymos as dm
from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE
from dymos.utils.lgl import lgl


@use_tempdirs
class TestPhaseInterp(unittest.TestCase):

    def test_linear_state(self):
        tx = dm.GaussLobatto(num_segments=8, order=5, compressed=True)
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)
        phase.add_state('x', fix_initial=True, fix_final=True)
        gd = tx.grid_data
        state_input_nodes = gd.node_ptau[gd.subset_node_indices['state_input']]
        expected = np.atleast_2d(100 * 0.5 * (state_input_nodes + 1)).T
        assert_near_equal(phase.interp('x', [0, 100]), expected)

    def test_quadratic_state(self):
        tx = dm.GaussLobatto(num_segments=8, order=5, compressed=True)
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)
        phase.add_state('x', fix_initial=True, fix_final=True)
        gd = tx.grid_data
        state_input_nodes = gd.node_ptau[gd.subset_node_indices['state_input']]
        expected = np.atleast_2d((2 * state_input_nodes) ** 2).T
        assert_near_equal(phase.interp('x', xs=[-2, 0, 2], ys=[4, 0, 4], kind='quadratic'), expected)

    def test_linear_control(self):
        tx = dm.GaussLobatto(num_segments=8, order=5, compressed=True)
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)
        phase.add_control('u', fix_initial=True, fix_final=True)
        gd = tx.grid_data
        input_nodes = gd.node_ptau[gd.subset_node_indices['control_input']]
        expected = np.atleast_2d(100 * 0.5 * (input_nodes + 1)).T
        assert_near_equal(phase.interp('u', [0, 100]), expected)

    def test_quadratic_control(self):
        tx = dm.GaussLobatto(num_segments=8, order=5, compressed=True)
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)
        phase.add_control('u', fix_initial=True, fix_final=True)
        gd = tx.grid_data
        input_nodes = gd.node_ptau[gd.subset_node_indices['control_input']]
        expected = np.atleast_2d((2 * input_nodes) ** 2).T
        assert_near_equal(phase.interp('u', xs=[-2, 0, 2], ys=[4, 0, 4], kind='quadratic'), expected)

    def test_polynomial_control(self):
        tx = dm.GaussLobatto(num_segments=8, order=5, compressed=True)
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)
        phase.add_control('u', fix_initial=True, fix_final=True, order=3, control_type='polynomial')
        xs = np.linspace(-10, 10, 100)
        ys = xs**3

        input_nodes, _ = lgl(4)
        expected = np.atleast_2d((10 * input_nodes) ** 3).T

        assert_near_equal(phase.interp('u', ys=ys, xs=xs, kind='cubic'), expected)

    def test_invalid_var(self):
        tx = dm.GaussLobatto(num_segments=8, order=5, compressed=True)
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)

        with self.assertRaises(ValueError) as e:
            phase.interp('x', [0, 100])

        expected = 'Could not find a state, control, or polynomial control named x to be ' \
                   'interpolated.\nPlease explicitly specify the node subset onto which this ' \
                   'value should be interpolated.'

        self.assertEqual(str(e.exception), expected)

    def test_invalid_var_explicit_nodes(self):
        tx = dm.GaussLobatto(num_segments=8, order=5, compressed=True)
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)
        phase.add_control('u', fix_initial=True, fix_final=True)
        gd = tx.grid_data
        state_input_nodes = gd.node_ptau[gd.subset_node_indices['control_input']]
        expected = np.atleast_2d(100 * 0.5 * (state_input_nodes + 1)).T
        assert_near_equal(phase.interp('foo', [0, 100], nodes='control_input'), expected)

    def test_invalid_ys(self):
        tx = dm.GaussLobatto(num_segments=8, order=5, compressed=True)
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)
        phase.add_control('u', fix_initial=True, fix_final=True)

        with self.assertRaises(ValueError) as e:
            phase.interp('foo', [0, 5, 10], nodes='control_input')

        expected = 'xs may only be unspecified when len(ys)=2'
        self.assertEqual(str(e.exception), expected)

    def test_invalid_kind(self):
        tx = dm.GaussLobatto(num_segments=8, order=5, compressed=True)
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)
        phase.add_control('u', fix_initial=True, fix_final=True)

        with self.assertRaises(ValueError) as e:
            phase.interp('u', [0, 5], nodes='control_input', kind='quadratic')

        expected = 'kind must be linear when xs is unspecified.'
        self.assertEqual(str(e.exception), expected)

    def test_invalid_xs(self):
        tx = dm.GaussLobatto(num_segments=8, order=5, compressed=True)
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)
        phase.add_control('u', fix_initial=True, fix_final=True)

        with self.assertRaises(ValueError) as e:
            phase.interp('u', xs=[[0, 1, 2], [3, 4, 5]], ys=[0, 1, 2, 3, 4, 5])

        expected = 'xs must be viewable as a 1D array'
        self.assertEqual(str(e.exception), expected)

    def test_underspecified(self):
        tx = dm.GaussLobatto(num_segments=8, order=5, compressed=True)
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)
        phase.add_control('u', fix_initial=True, fix_final=True)

        with self.assertRaises(ValueError) as e:
            phase.interp(ys=[0, 5])

        expected = 'nodes for interpolation were not specified but the name of the variable ' \
                   'to be interpolated was not provided.\nPlease specify the name of the ' \
                   'interpolated variable or a node subset.'

        self.assertEqual(str(e.exception), expected)

    def test_nodes_sequence(self):
        tx = dm.GaussLobatto(num_segments=6, order=3, compressed=True)
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)

        interp_vals = phase.interp(ys=[0, 5], nodes=np.linspace(-1, 1, 100))
        check_vals = np.atleast_2d(np.linspace(0, 5, 100)).T

        assert_near_equal(interp_vals, check_vals)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import numpy as np


class Track:
    def __init__(self, segments):
        self.segments = segments
        self.cornerspeeds = np.zeros(len(segments), dtype=complex)

    def get_segment_type(self, num):
        return self.segments[num][0]

    def get_segment_length(self, num):
        return self.segments[num][1]

    def get_segment_radius(self, num):
        return self.segments[num][2]

    def get_corner_direction(self, num):
        return self.segments[num][3]

    def set_corner_speed(self, num, speed):
        self.cornerspeeds[num] = speed

    def get_corner_speed(self, num):
        return self.cornerspeeds[num]

    def get_total_length(self):
        length = 0
        for i in range(len(self.segments)):
            if self.segments[i][0] == 0:
                # straight
                length += self.get_segment_length(i)
            else:
                # corner
                length += self.get_segment_length(i) * self.get_segment_radius(i)
        return length



import openmdao.api as om
import dymos as dm
import unittest
import os

try:
    import matplotlib  # noqa: F401
    SHOW_PLOTS = True
except ImportError:
    SHOW_PLOTS = False

from openmdao.utils.assert_utils import assert_near_equal, assert_check_partials
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse

from dymos.examples.goddard_rocket_problem import RocketODE


@require_pyoptsparse(optimizer='IPOPT')
def goddard_rocket_direct_collocation(grid_type='lgl'):

    optimizer = 'IPOPT'
    p = om.Problem(model=om.Group())
    p.driver = om.pyOptSparseDriver()
    p.driver.options['optimizer'] = optimizer
    p.driver.declare_coloring()
    if optimizer == 'SNOPT':
        p.driver.opt_settings['Major iterations limit'] = 1000
        p.driver.opt_settings['Major feasibility tolerance'] = 1.0E-6
        p.driver.opt_settings['Major optimality tolerance'] = 1.0E-3
        p.driver.opt_settings['iSumm'] = 6

    t = dm.Birkhoff(num_nodes=150, grid_type=grid_type)

    traj = p.model.add_subsystem('traj', dm.Trajectory())

    phase = traj.add_phase('phase0', dm.Phase(ode_class=RocketODE, transcription=t))

    phase.set_time_options(fix_initial=True, fix_duration=False, duration_bounds=(20, 80), units='s',
                           duration_ref=100)

    phase.add_state('h', fix_initial=True, fix_final=False, rate_source='h_dot', units='ft', lower=0, ref=20000)
    phase.add_state('v', fix_initial=True, fix_final=False, rate_source='v_dot', units='ft/s', ref=1000, lower=0)
    phase.add_state('m', fix_initial=True, fix_final=True, rate_source='m_dot', units='slug',
                    lower=1e-3, ref=3, upper=3.0)

    phase.add_control('T', units='lbf', lower=0, upper=200, ref=200)

    # Minimize the control effort
    phase.add_objective('h', loc='final', scaler=-1)

    p.model.linear_solver = om.DirectSolver()

    p.setup(check=True, force_alloc_complex=True)

    phase.set_time_val(initial=0.0, duration=40.0)
    phase.set_state_val('h', [0.0, 10000.0])
    phase.set_state_val('v', [0.0, 0.0])
    phase.set_state_val('m', [3.0, 1.0])
    phase.set_control_val('T', [0.0, 0.0])

    return p


@use_tempdirs
class TestGoddardRocketProblem(unittest.TestCase):

    @classmethod
    def tearDownClass(cls):
        for filename in ['total_coloring.pkl', 'SLSQP.out', 'SNOPT_print.out']:
            if os.path.exists(filename):
                os.remove(filename)

    @staticmethod
    def _assert_results(p, tol=1.0E-4):
        h = p.get_val('traj.phase0.timeseries.h')
        assert_near_equal(h[-1], 18745.0, tolerance=tol)

    def test_goddard_rocket_problem_lgl(self):
        p = goddard_rocket_direct_collocation(grid_type='lgl')
        dm.run_problem(p)
        if SHOW_PLOTS:
            t = p.get_val('traj.phase0.timeseries.time')
            h = p.get_val('traj.phase0.timeseries.h')
            v = p.get_val('traj.phase0.timeseries.v')
            m = p.get_val('traj.phase0.timeseries.m')
            T = p.get_val('traj.phase0.timeseries.T')

            import matplotlib.pyplot as plt
            fig, axs = plt.subplots(2, 2)
            axs[0, 0].plot(t, h)
            axs[0, 0].set_xlabel('time (s)')
            axs[0, 0].set_ylabel('altitude (ft)')
            axs[0, 1].plot(t, v)
            axs[0, 1].set_xlabel('time (s)')
            axs[0, 1].set_ylabel('velocity (ft/s)')
            axs[1, 0].plot(t, m)
            axs[1, 0].set_xlabel('time (s)')
            axs[1, 0].set_ylabel('mass (slug)')
            axs[1, 1].plot(t, T)
            axs[1, 1].set_xlabel('time (s)')
            axs[1, 1].set_ylabel('thrust (lb)')
            plt.show()

        self._assert_results(p)

    def test_goddard_rocket_problem_cgl(self):
        p = goddard_rocket_direct_collocation(grid_type='cgl')
        dm.run_problem(p)
        if SHOW_PLOTS:
            t = p.get_val('traj.phase0.timeseries.time')
            h = p.get_val('traj.phase0.timeseries.h')
            v = p.get_val('traj.phase0.timeseries.v')
            m = p.get_val('traj.phase0.timeseries.m')
            T = p.get_val('traj.phase0.timeseries.T')

            import matplotlib.pyplot as plt
            fig, axs = plt.subplots(2, 2)
            axs[0, 0].plot(t, h)
            axs[0, 1].plot(t, v)
            axs[1, 0].plot(t, m)
            axs[1, 1].plot(t, T)
            plt.show()
        self._assert_results(p)

    def test_check_partials(self):
        p = goddard_rocket_direct_collocation(grid_type='cgl')
        p.run_model()
        cpd = p.check_partials(method='cs', out_stream=None)
        assert_check_partials(cpd)


if __name__ == "__main__":
    unittest.main()



import os
import unittest

try:
    import matplotlib
except ImportError:
    matplotlib = None

from openmdao.utils.testing_utils import use_tempdirs
from dymos.utils.misc import om_version


@use_tempdirs
class TestBrachistochroneRecordingExample(unittest.TestCase):

    @classmethod
    def tearDownClass(cls):
        for filename in ['brachistochrone_solution.db']:
            if os.path.exists(filename):
                os.remove(filename)

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_brachistochrone_recording(self):
        import matplotlib
        matplotlib.use('Agg')
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm
        from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE

        p = om.Problem(model=om.Group())
        p.driver = om.ScipyOptimizeDriver()

        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=dm.GaussLobatto(num_segments=10))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)
        phase.add_state('y', fix_initial=True, fix_final=True)
        phase.add_state('v', fix_initial=True)

        phase.add_control('theta', units='deg',
                          rate_continuity=False, lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()

        p.setup()

        phase.set_time_val(initial=0.0, duration=2.0)

        phase.set_state_val('x', [0, 10], units='m')
        phase.set_state_val('y', [10, 5], units='m')
        phase.set_state_val('v', [0, 9.9], units='m/s')

        phase.set_control_val('theta', [5, 100.5], units='deg')

        # Solve for the optimal trajectory
        dm.run_problem(p)

        # Test the results
        assert_near_equal(p.get_val('phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-3)

        sol_db = 'dymos_solution.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = p.get_outputs_dir() / sol_db
        case = om.CaseReader(sol_db).get_case('final')

        assert_near_equal(p['phase0.control_values:theta'],
                          case.get_val('phase0.timeseries.theta'))


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



from .collocation_comp import CollocationComp
from .state_interp_comp import StateInterpComp
from .state_independents import StateIndependentsComp
from .control_endpoint_defect_comp import ControlEndpointDefectComp
from .gauss_lobatto_interleave_comp import GaussLobattoInterleaveComp
from .birkhoff_defect_comp import BirkhoffDefectComp
from .birkhoff_iter_group import BirkhoffIterGroup
from .birkhoff_boundary_group import BirkhoffBoundaryGroup



"""Define the CollocationComp class."""
import numpy as np

import openmdao.api as om

from ...grid_data import GridData
from ....utils.misc import get_rate_units
from ...._options import options as dymos_options


class CollocationComp(om.ExplicitComponent):
    """
    Class definiton for the Collocationcomp.

    CollocationComp computes the generalized defect of a segment for implicit collocation.
    The defect is the interpolated state derivative at the collocation nodes minus
    the computed state derivative at the collocation nodes.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._no_check_partials = not dymos_options['include_check_partials']

    def initialize(self):
        """
        Declare component options.
        """
        self.options.declare(
            'grid_data', types=GridData,
            desc='Container object for grid info')

        self.options.declare(
            'state_options', types=dict,
            desc='Dictionary of state names/options for the phase')

        self.options.declare(
            'time_units', default=None, allow_none=True, types=str,
            desc='Units of time')

    def configure_io(self):
        """
        I/O creation is delayed until configure so we can determine shape and units.
        """
        gd = self.options['grid_data']
        num_col_nodes = gd.subset_num_nodes['col']
        time_units = self.options['time_units']
        state_options = self.options['state_options']

        self.add_input('dt_dstau', units=time_units, shape=(num_col_nodes,))

        self.var_names = var_names = {}
        for state_name in state_options:
            var_names[state_name] = {
                'f_approx': f'f_approx:{state_name}',
                'f_computed': f'f_computed:{state_name}',
                'defect': f'defects:{state_name}',
            }

        for state_name, options in state_options.items():
            shape = options['shape']
            units = options['units']

            rate_units = get_rate_units(units, time_units)

            var_names = self.var_names[state_name]

            self.add_input(
                name=var_names['f_approx'],
                shape=(num_col_nodes,) + shape,
                desc=f'Estimated derivative of state {state_name} at the collocation nodes',
                units=rate_units)

            self.add_input(
                name=var_names['f_computed'],
                shape=(num_col_nodes,) + shape,
                desc=f'Computed derivative of state {state_name} at the collocation nodes',
                units=rate_units)

            self.add_output(
                name=var_names['defect'],
                shape=(num_col_nodes,) + shape,
                desc=f'Interior defects of state {state_name}',
                units=units)

            if 'defect_ref' in options and options['defect_ref'] is not None:
                defect_ref = options['defect_ref']
            elif 'defect_scaler' in options and options['defect_scaler'] is not None:
                defect_ref = np.divide(1.0, options['defect_scaler'])
            else:
                if 'ref' in options and options['ref'] is not None:
                    defect_ref = options['ref']
                elif 'scaler' in options and options['scaler'] is not None:
                    defect_ref = np.divide(1.0, options['scaler'])
                else:
                    defect_ref = 1.0

            if not np.isscalar(defect_ref):
                defect_ref = np.asarray(defect_ref)
                if defect_ref.shape == shape:
                    defect_ref = np.tile(defect_ref.flatten(), num_col_nodes)
                else:
                    raise ValueError('array-valued scaler/ref must length equal to state-size')

            if not options['solve_segments']:
                self.add_constraint(name=var_names['defect'],
                                    equals=0.0,
                                    ref=defect_ref)

        # Setup partials
        num_col_nodes = self.options['grid_data'].subset_num_nodes['col']
        state_options = self.options['state_options']

        for state_name, options in state_options.items():
            shape = options['shape']
            size = np.prod(shape)

            r = np.arange(num_col_nodes * size)

            var_names = self.var_names[state_name]

            self.declare_partials(of=var_names['defect'],
                                  wrt=var_names['f_approx'],
                                  rows=r, cols=r)

            self.declare_partials(of=var_names['defect'],
                                  wrt=var_names['f_computed'],
                                  rows=r, cols=r)

            c = np.repeat(np.arange(num_col_nodes), size)
            self.declare_partials(of=var_names['defect'],
                                  wrt='dt_dstau',
                                  rows=r, cols=c)

    def compute(self, inputs, outputs):
        """
        Compute collocation defects.

        Parameters
        ----------
        inputs : `Vector`
            `Vector` containing inputs.
        outputs : `Vector`
            `Vector` containing outputs.
        """
        state_options = self.options['state_options']
        dt_dstau = inputs['dt_dstau']

        for state_name in state_options:
            var_names = self.var_names[state_name]

            f_approx = inputs[var_names['f_approx']]
            f_computed = inputs[var_names['f_computed']]

            outputs[var_names['defect']] = ((f_approx - f_computed).T * dt_dstau).T

    def compute_partials(self, inputs, partials):
        """
        Compute sub-jacobian parts. The model is assumed to be in an unscaled state.

        Parameters
        ----------
        inputs : Vector
            Unscaled, dimensional input variables read via inputs[key].
        partials : Jacobian
            Subjac components written to partials[output_name, input_name].
        """
        dt_dstau = inputs['dt_dstau']
        for state_name, options in self.options['state_options'].items():
            size = np.prod(options['shape'])
            var_names = self.var_names[state_name]
            f_approx = inputs[var_names['f_approx']]
            f_computed = inputs[var_names['f_computed']]

            k = np.repeat(dt_dstau, size)

            partials[var_names['defect'], var_names['f_approx']] = k
            partials[var_names['defect'], var_names['f_computed']] = -k
            partials[var_names['defect'], 'dt_dstau'] = (f_approx - f_computed).ravel()



import unittest

import openmdao.api as om
import dymos as dm

from openmdao.utils.testing_utils import use_tempdirs
from openmdao.utils.assert_utils import assert_near_equal
from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE


@use_tempdirs
class TestBrachistochroneExprPathConstraint(unittest.TestCase):

    def _make_problem(self, tx):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=tx)

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.set_state_options('x', fix_initial=True)
        phase.set_state_options('y', fix_initial=True)
        phase.set_state_options('v', fix_initial=True)

        phase.add_control('theta', continuity=True, rate_continuity=True, opt=True,
                          units='deg', lower=0.01, upper=179.9, ref=1, ref0=0)

        phase.add_boundary_constraint('x', loc='final', equals=10.0)
        phase.add_path_constraint('pc = y-x/2-1', lower=0.0)

        phase.add_parameter('g', opt=False, units='m/s**2', val=9.80665, shape=None, include_timeseries=True)

        phase.add_objective('time_phase', loc='final', scaler=10)

        p.model.options['assembled_jac_type'] = 'csc'
        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True)

        phase.set_time_val(initial=0.0, duration=2.0)

        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])

        phase.set_control_val('theta', [5, 100])
        phase.set_parameter_val('g', 9.80665)
        return p

    def test_brachistochrone_expr_path_constraint(self):
        prob = self._make_problem(tx=dm.Radau(num_segments=5, order=3, compressed=True))
        prob.run_driver()
        yf = prob.get_val('phase0.timeseries.y')[-1]

        assert_near_equal(yf, 6)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import os
import pathlib
import unittest

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs

import dymos as dm
from dymos.utils.misc import om_version
from dymos.examples.aircraft_steady_flight.aircraft_ode import AircraftODE

optimizer = os.environ.get('DYMOS_DEFAULT_OPT', 'SLSQP')


@use_tempdirs
class TestAircraftCruise(unittest.TestCase):

    def test_cruise_results_gl(self):
        p = om.Problem()
        if optimizer == 'SNOPT':
            p.driver = om.pyOptSparseDriver()
            p.driver.options['optimizer'] = optimizer
            p.driver.declare_coloring()
            p.driver.opt_settings['Major iterations limit'] = 100
            p.driver.opt_settings['Major step limit'] = 0.05
            p.driver.opt_settings['Major feasibility tolerance'] = 1.0E-6
            p.driver.opt_settings['Major optimality tolerance'] = 1.0E-6
            p.driver.opt_settings["Linesearch tolerance"] = 0.10
            p.driver.opt_settings['iSumm'] = 6
            p.driver.opt_settings['Verify level'] = 3
        else:
            p.driver = om.ScipyOptimizeDriver()
            p.driver.declare_coloring()

        # Pass Reference Area from an external source
        assumptions = p.model.add_subsystem('assumptions', om.IndepVarComp())
        assumptions.add_output('S', val=427.8, units='m**2')
        assumptions.add_output('mass_empty', val=1.0, units='kg')
        assumptions.add_output('mass_payload', val=1.0, units='kg')

        transcription = dm.GaussLobatto(num_segments=1,
                                        order=13,
                                        compressed=False)
        phase = dm.Phase(ode_class=AircraftODE, transcription=transcription)
        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True,
                               duration_bounds=(3600, 3600),
                               duration_ref=3600)

        phase.add_state('range', units='km', fix_initial=True, fix_final=False, scaler=0.01,
                        rate_source='range_rate_comp.dXdt:range',
                        defect_scaler=0.01)
        phase.add_state('mass_fuel', units='kg', fix_final=True, upper=20000.0, lower=0.0,
                        rate_source='propulsion.dXdt:mass_fuel',
                        scaler=1.0E-4, defect_scaler=1.0E-2)
        phase.add_state('alt',
                        rate_source='climb_rate',
                        units='km', fix_initial=True)

        phase.add_control('mach', targets=['tas_comp.mach', 'aero.mach'], units=None, opt=False)

        phase.add_control('climb_rate', targets=['gam_comp.climb_rate'], units='m/s', opt=False)

        phase.add_parameter('S',
                            targets=['aero.S', 'flight_equilibrium.S', 'propulsion.S'],
                            units='m**2')

        phase.add_parameter('mass_empty', targets=['mass_comp.mass_empty'], units='kg')
        phase.add_parameter('mass_payload', targets=['mass_comp.mass_payload'], units='kg')

        phase.add_path_constraint('propulsion.tau', lower=0.01, upper=1.0)

        phase.add_timeseries_output('tas_comp.TAS')

        p.model.connect('assumptions.S', 'phase0.parameters:S')
        p.model.connect('assumptions.mass_empty', 'phase0.parameters:mass_empty')
        p.model.connect('assumptions.mass_payload', 'phase0.parameters:mass_payload')

        phase.add_objective('time', loc='final', ref=3600)

        p.model.linear_solver = om.DirectSolver()

        p.setup()

        phase.set_time_val(initial=0.0, duration=3600)
        phase.set_state_val('range', (0, 1296.4))
        phase.set_state_val('mass_fuel', (12236.594555, 1))
        phase.set_state_val('alt', 5.0)
        phase.set_control_val('mach', 0.8)
        phase.set_control_val('climb_rate', 0.0)

        p['assumptions.S'] = 427.8
        p['assumptions.mass_empty'] = 0.15E6
        p['assumptions.mass_payload'] = 84.02869 * 400

        dm.run_problem(p)

        time = p.get_val('phase0.timeseries.time')
        tas = p.get_val('phase0.timeseries.TAS', units='km/s')
        range = p.get_val('phase0.timeseries.range')

        assert_near_equal(range, tas*time, tolerance=1.0E-4)

        exp_out = phase.simulate()

        time = exp_out.get_val('phase0.timeseries.time')
        tas = exp_out.get_val('phase0.timeseries.TAS', units='km/s')
        range = exp_out.get_val('phase0.timeseries.range')

        assert_near_equal(range, tas*time, tolerance=1.0E-4)

    def test_cruise_results_radau(self):
        p = om.Problem(model=om.Group())
        if optimizer == 'SNOPT':
            p.driver = om.pyOptSparseDriver()
            p.driver.options['optimizer'] = optimizer
            p.driver.declare_coloring()
            p.driver.opt_settings['Major iterations limit'] = 100
            p.driver.opt_settings['Major step limit'] = 0.05
            p.driver.opt_settings['Major feasibility tolerance'] = 1.0E-6
            p.driver.opt_settings['Major optimality tolerance'] = 1.0E-6
            p.driver.opt_settings["Linesearch tolerance"] = 0.10
            p.driver.opt_settings['iSumm'] = 6
            p.driver.opt_settings['Verify level'] = 3
        else:
            p.driver = om.ScipyOptimizeDriver()
            p.driver.declare_coloring()

        # Pass Reference Area from an external source
        assumptions = p.model.add_subsystem('assumptions', om.IndepVarComp())
        assumptions.add_output('S', val=427.8, units='m**2')
        assumptions.add_output('mass_empty', val=1.0, units='kg')
        assumptions.add_output('mass_payload', val=1.0, units='kg')

        transcription = dm.Radau(num_segments=1, order=13, compressed=False)
        phase = dm.Phase(ode_class=AircraftODE, transcription=transcription)
        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True,
                               duration_bounds=(3600, 3600),
                               duration_ref=3600)

        phase.add_state('range', units='km', fix_initial=True, fix_final=False, scaler=0.01,
                        rate_source='range_rate_comp.dXdt:range',
                        defect_scaler=0.01)
        phase.add_state('mass_fuel', units='kg', fix_final=True, upper=20000.0, lower=0.0,
                        rate_source='propulsion.dXdt:mass_fuel',
                        scaler=1.0E-4, defect_scaler=1.0E-2)
        phase.add_state('alt',
                        rate_source='climb_rate',
                        units='km', fix_initial=True)

        phase.add_control('mach', targets=['tas_comp.mach', 'aero.mach'], opt=False)

        phase.add_control('climb_rate', targets=['gam_comp.climb_rate'], units='m/s', opt=False)

        phase.add_parameter('S',
                            targets=['aero.S', 'flight_equilibrium.S', 'propulsion.S'],
                            units='m**2')

        phase.add_parameter('mass_empty', targets=['mass_comp.mass_empty'], units='kg')
        phase.add_parameter('mass_payload', targets=['mass_comp.mass_payload'], units='kg')

        phase.add_path_constraint('propulsion.tau', lower=0.01, upper=1.0)

        phase.add_timeseries_output('tas_comp.TAS')

        p.model.connect('assumptions.S', 'phase0.parameters:S')
        p.model.connect('assumptions.mass_empty', 'phase0.parameters:mass_empty')
        p.model.connect('assumptions.mass_payload', 'phase0.parameters:mass_payload')

        phase.add_objective('time', loc='final', ref=3600)

        p.model.linear_solver = om.DirectSolver()

        p.setup()

        phase.set_time_val(initial=0.0, duration=3600)
        phase.set_state_val('range', (0, 1296.4))
        phase.set_state_val('mass_fuel', (12236.594555, 0))
        phase.set_state_val('alt', 5.0)
        phase.set_control_val('mach', 0.8)
        phase.set_control_val('climb_rate', 0.0)

        p['assumptions.S'] = 427.8
        p['assumptions.mass_empty'] = 0.15E6
        p['assumptions.mass_payload'] = 84.02869 * 400

        dm.run_problem(p)

        time = p.get_val('phase0.timeseries.time')
        tas = p.get_val('phase0.timeseries.TAS', units='km/s')
        range = p.get_val('phase0.timeseries.range')

        assert_near_equal(range, tas*time, tolerance=1.0E-4)

        exp_out = phase.simulate()

        time = exp_out.get_val('phase0.timeseries.time')
        tas = exp_out.get_val('phase0.timeseries.TAS', units='km/s')
        range = exp_out.get_val('phase0.timeseries.range')

        assert_near_equal(range, tas*time, tolerance=1.0E-4)

    def test_cruise_results_birkhoff(self):
        p = om.Problem(model=om.Group())
        optimizer = 'SLSQP'
        if optimizer == 'SNOPT':
            p.driver = om.pyOptSparseDriver()
            p.driver.options['optimizer'] = optimizer
            p.driver.declare_coloring()
            p.driver.opt_settings['Major iterations limit'] = 100
            p.driver.opt_settings['Major step limit'] = 0.05
            p.driver.opt_settings['Major feasibility tolerance'] = 1.0E-6
            p.driver.opt_settings['Major optimality tolerance'] = 1.0E-6
            p.driver.opt_settings["Linesearch tolerance"] = 0.10
            p.driver.opt_settings['iSumm'] = 6
            p.driver.opt_settings['Verify level'] = 3
        else:
            p.driver = om.ScipyOptimizeDriver()
            p.driver.declare_coloring()

        # Pass Reference Area from an external source
        assumptions = p.model.add_subsystem('assumptions', om.IndepVarComp())
        assumptions.add_output('S', val=427.8, units='m**2')
        assumptions.add_output('mass_empty', val=1.0, units='kg')
        assumptions.add_output('mass_payload', val=1.0, units='kg')

        transcription = dm.Birkhoff(num_nodes=50, grid_type='lgl', solve_segments='forward')
        phase = dm.Phase(ode_class=AircraftODE, transcription=transcription)
        traj = dm.Trajectory()
        traj.add_phase('phase0', phase)
        p.model.add_subsystem('traj', traj)

        phase.set_time_options(fix_initial=True,
                               duration_bounds=(3600, 3600),
                               duration_ref=3600)

        phase.add_state('range', units='km', fix_initial=True, fix_final=False, scaler=0.01,
                        rate_source='range_rate_comp.dXdt:range',
                        defect_scaler=0.01)
        phase.add_state('mass_fuel', units='kg', fix_final=False, upper=20000.0, lower=0.0,
                        rate_source='propulsion.dXdt:mass_fuel',
                        scaler=1.0E-4, defect_scaler=1.0E-2)
        phase.add_state('alt',
                        rate_source='climb_rate',
                        units='km', fix_initial=True)

        phase.add_control('mach', targets=['tas_comp.mach', 'aero.mach'], opt=False)

        phase.add_control('climb_rate', targets=['gam_comp.climb_rate'], units='m/s', opt=False)

        phase.add_parameter('S',
                            targets=['aero.S', 'flight_equilibrium.S', 'propulsion.S'],
                            units='m**2')

        phase.add_parameter('mass_empty', targets=['mass_comp.mass_empty'], units='kg')
        phase.add_parameter('mass_payload', targets=['mass_comp.mass_payload'], units='kg')

        phase.add_path_constraint('propulsion.tau', lower=0.01, upper=1.0)
        phase.add_boundary_constraint('mass_fuel', loc='final', equals=0.)
        phase.add_boundary_constraint('range_rate_comp.dXdt:range', loc='final', lower=0.)
        phase.add_boundary_constraint('range_rate_comp.dXdt:range', loc='initial', lower=0.)

        phase.add_timeseries_output('tas_comp.TAS')

        p.model.connect('assumptions.S', 'traj.phase0.parameters:S')
        p.model.connect('assumptions.mass_empty', 'traj.phase0.parameters:mass_empty')
        p.model.connect('assumptions.mass_payload', 'traj.phase0.parameters:mass_payload')

        phase.add_objective('time', loc='final', ref=3600)

        p.model.linear_solver = om.DirectSolver()

        p.setup()

        phase.set_time_val(initial=0.0, duration=3600.0)
        phase.set_state_val('range', [0.0, 500.0])
        phase.set_state_val('mass_fuel', [10000.0, 1.0])
        phase.set_state_val('alt', [10.0, 10.0])
        phase.set_control_val('mach', 0.8)
        phase.set_control_val('climb_rate', 0.0)

        p.set_val('assumptions.S', 427.8)
        p.set_val('assumptions.mass_empty', 0.15E6)
        p.set_val('assumptions.mass_payload', 84.02869 * 400)

        dm.run_problem(p, simulate=True, make_plots=True)

        time = p.get_val('traj.phase0.timeseries.time')
        tas = p.get_val('traj.phase0.timeseries.TAS', units='km/s')
        range = p.get_val('traj.phase0.timeseries.range')
        mass_fuel = p.get_val('traj.phase0.timeseries.mass_fuel')

        assert_near_equal(range, tas*time, tolerance=1.0E-4)
        assert_near_equal(mass_fuel[-1, ...], 0.0, tolerance=1.0E-4)

        if om_version()[0] > (3, 34, 2):
            sim_out_dir = traj.sim_prob.get_outputs_dir()
        else:
            sim_out_dir = pathlib.Path.cwd()
        case = om.CaseReader(sim_out_dir / 'dymos_simulation.db').get_case('final')

        time = case.get_val('traj.phase0.timeseries.time')
        tas = case.get_val('traj.phase0.timeseries.TAS', units='km/s')
        range = case.get_val('traj.phase0.timeseries.range')
        mass_fuel = case.get_val('traj.phase0.timeseries.mass_fuel')

        assert_near_equal(range, tas*time, tolerance=1.0E-4)
        assert_near_equal(mass_fuel[-1, ...], 0.0, tolerance=1.0E-4)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest

import numpy as np

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from dymos.utils.testing_utils import assert_check_partials

import dymos as dm
from dymos.transcriptions.pseudospectral.components import ControlEndpointDefectComp
from dymos.transcriptions.grid_data import GridData


class TestControlEndpointDefectComp(unittest.TestCase):

    def setUp(self):
        dm.options['include_check_partials'] = True
        gd = GridData(num_segments=2, segment_ends=np.array([0., 2., 4.]),
                      transcription='radau-ps', transcription_order=3)

        self.gd = gd

        self.p = om.Problem(model=om.Group())

        control_opts = {'u': {'units': 'm', 'shape': (1,), 'dynamic': True, 'opt': True},
                        'v': {'units': 'm', 'shape': (3, 2), 'dynamic': True, 'opt': True}}

        indep_comp = om.IndepVarComp()
        self.p.model.add_subsystem('indep', indep_comp, promotes=['*'])

        indep_comp.add_output('controls:u',
                              val=np.zeros((gd.subset_num_nodes['all'], 1)), units='m')

        indep_comp.add_output('controls:v',
                              val=np.zeros((gd.subset_num_nodes['all'], 3, 2)), units='m')

        self.p.model.add_subsystem('endpoint_defect_comp',
                                   subsys=ControlEndpointDefectComp(grid_data=gd,
                                                                    control_options=control_opts))

        self.p.model.connect('controls:u', 'endpoint_defect_comp.controls:u')
        self.p.model.connect('controls:v', 'endpoint_defect_comp.controls:v')

        self.p.setup(force_alloc_complex=True)

        self.p['controls:u'] = np.random.random((gd.subset_num_nodes['all'], 1))
        self.p['controls:v'] = np.random.random((gd.subset_num_nodes['all'], 3, 2))

        self.p.run_model()

    def tearDown(self):
        dm.options['include_check_partials'] = False

    def test_results(self):

        u_coefs = np.polyfit(self.gd.node_ptau[-4:-1], self.p['controls:u'][-4:-1], deg=2)
        u_poly = np.poly1d(u_coefs.ravel())
        u_interp = u_poly(1.0)
        u_given = self.p['controls:u'][-1]
        assert_near_equal(np.ravel(self.p['endpoint_defect_comp.control_endpoint_defects:u']),
                          np.ravel(u_given - u_interp),
                          tolerance=1.0E-9)

    def test_partials(self):
        cpd = self.p.check_partials(compact_print=False, method='cs', out_stream=None)
        assert_check_partials(cpd)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.mpi import MPI
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse

from dymos.examples.finite_burn_orbit_raise.finite_burn_orbit_raise_problem import two_burn_orbit_raise_problem
from dymos.utils.misc import om_version


@require_pyoptsparse(optimizer='IPOPT')
@unittest.skipUnless(MPI, "MPI is required.")
@use_tempdirs
class TestExampleTwoBurnOrbitRaiseMPI(unittest.TestCase):
    N_PROCS = 3

    def test_ex_two_burn_orbit_raise_mpi(self):
        optimizer = 'IPOPT'

        CONNECTED = False

        p = two_burn_orbit_raise_problem(transcription='radau', transcription_order=3,
                                         compressed=False, optimizer=optimizer, simulate=True,
                                         connected=CONNECTED, show_output=False)

        sol_db = 'dymos_solution.db'
        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = p.get_outputs_dir() / sol_db
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        sol_case = om.CaseReader(sol_db).get_case('final')
        sim_case = om.CaseReader(sim_db).get_case('final')

        # The last phase in this case is run in reverse time if CONNECTED=True,
        # so grab the correct index to test the resulting delta-V.
        end_idx = 0 if CONNECTED else -1

        assert_near_equal(sol_case.get_val('traj.burn2.timeseries.deltav')[end_idx], 0.3995, tolerance=2.0E-3)
        assert_near_equal(sim_case.get_val('traj.burn2.timeseries.deltav')[end_idx], 0.3995, tolerance=2.0E-3)

    def test_ex_two_burn_orbit_raise_connected_mpi(self):
        optimizer = 'IPOPT'

        CONNECTED = True

        p = two_burn_orbit_raise_problem(transcription='radau', transcription_order=3,
                                         compressed=False, optimizer=optimizer, simulate=True,
                                         connected=CONNECTED, show_output=False)

        sol_db = 'dymos_solution.db'
        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = p.get_outputs_dir() / sol_db
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        sol_case = om.CaseReader(sol_db).get_case('final')
        sim_case = om.CaseReader(sim_db).get_case('final')

        # The last phase in this case is run in reverse time if CONNECTED=True,
        # so grab the correct index to test the resulting delta-V.
        end_idx = 0 if CONNECTED else -1

        assert_near_equal(sol_case.get_val('traj.burn2.timeseries.deltav')[end_idx], 0.3995, tolerance=2.0E-3)
        assert_near_equal(sim_case.get_val('traj.burn2.timeseries.deltav')[end_idx], 0.3995, tolerance=2.0E-3)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import numpy as np
import openmdao.api as om

from ...grid_data import GridData
from ....utils.ode_utils import _make_ode_system
from dymos._options import options as dymos_options


class BirkhoffBoundaryMuxComp(om.ExplicitComponent):
    """
    Class definition of the BirtkhoffBoundaryMuxComp.

    This component takes the initial and final values of states
    and muxes them into a single output.

    For a state of a given `shape` in a phase with `num_seg` segments,
    the shape of `initial_states:{state_name}` and `final_states:{state_name}`
    are both `(num_seg,) + shape` and the shape of the resulting
    `states:{state_name}` is `(2,) + shape`.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional phase arguments.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._io_names = {}
        self._no_check_partials = not dymos_options['include_check_partials']

    def configure_io(self, state_options):
        """
        I/O creation is delayed until configure so that we can determine shape and units for the states.

        Parameters
        ----------
        state_options : StateOptionsDictionary
            The phase object to which this transcription instance applies.
        """
        self._io_names = {}
        for state_name, options in state_options.items():
            shape = options['shape']
            size = np.prod(shape, dtype=int)
            units = options['units']
            self._io_names[state_name] = {'initial': f'initial_states:{state_name}',
                                          'final': f'final_states:{state_name}',
                                          'boundary': state_name}

            iname = self._io_names[state_name]['initial']
            fname = self._io_names[state_name]['final']
            bname = self._io_names[state_name]['boundary']

            self.add_input(iname, shape=shape, units=units)
            self.add_input(fname, shape=shape, units=units)
            self.add_output(bname, shape=(2,) + shape, units=units)

            ar = np.arange(size, dtype=int)

            self.declare_partials(of=bname, wrt=iname, rows=ar, cols=ar, val=1.0)
            self.declare_partials(of=bname, wrt=fname, rows=ar + size, cols=ar, val=1.0)

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        """
        Compute component outputs.

        Parameters
        ----------
        inputs : Vector
            Unscaled, dimensional input variables read via inputs[key].
        outputs : Vector
            Unscaled, dimensional output variables read via outputs[key].
        discrete_inputs : dict or None
            If not None, dict containing discrete input values.
        discrete_outputs : dict or None
            If not None, dict containing discrete output values.
        """
        for state_name, io_names in self._io_names.items():
            outputs[io_names['boundary']][0] = inputs[io_names['initial']]
            outputs[io_names['boundary']][1] = inputs[io_names['final']]


class BirkhoffBoundaryGroup(om.Group):
    """
    Class definition for the BirkhoffBoundaryGroup.

    This group accepts values for initial and final times, states, controls, and parameters
    and evaluates the ODE with those in order to compute the boundary values and
    objectives.

    Note that in the Birkhoff transcription, the initial and final state values are
    decoupled from the initial and final states in the interpolating polynomial.

    Dymos uses the Birkhoff LGL or CGL approaches so that the control values are provided
    at the endpoints of the phase without the need for extrapolation (unlike the classical
    Radau approach in Dymos)

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._no_check_partials = not dymos_options['include_check_partials']

    def initialize(self):
        """
        Declare group options.
        """
        self.options.declare('grid_data', types=GridData, desc='Container object for grid info.')
        self.options.declare('ode_class', default=None,
                             desc='Callable that instantiates the ODE system.',
                             recordable=False)
        self.options.declare('ode_init_kwargs', types=dict, default={}, recordable=False,
                             desc='Keyword arguments provided when initializing the ODE System')
        self.options.declare('calc_exprs', types=dict, default={},
                             desc='ODE Expressions from the Phase')
        self.options.declare('parameter_options', types=dict, default={},
                             desc='parameter options inherited from the phase')

    def setup(self):
        """
        Define the structure of the control group.
        """
        ode_class = self.options['ode_class']
        ode_init_kwargs = self.options['ode_init_kwargs']

        self.add_subsystem('boundary_mux', subsys=BirkhoffBoundaryMuxComp(),
                           promotes_inputs=['*'], promotes_outputs=['*'])

        ode = _make_ode_system(ode_class=ode_class,
                               num_nodes=2,
                               ode_init_kwargs=ode_init_kwargs,
                               calc_exprs=self.options['calc_exprs'],
                               parameter_options=self.options['parameter_options'])

        self.add_subsystem('boundary_ode', ode,
                           promotes_inputs=['*'], promotes_outputs=['*'])

    def configure_io(self, phase):
        """
        I/O creation is delayed until configure so that we can determine shape and units for the states.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        self._get_subsystem('boundary_mux').configure_io(state_options=phase.state_options)

        for state_name, options in phase.state_options.items():
            for tgt in options['targets']:
                self.promotes('boundary_ode', inputs=[(tgt, state_name)])



import numpy as np

import openmdao.api as om


class ThrottleComp(om.ExplicitComponent):
    """ Compute 'tau' (throttle parameter) which is the ratio of the current thrust
        (as determined to provie flight equilibrium) with the maximum thrust given
        the current aircraft state.

    """
    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # Inputs
        self.add_input(name='thrust', shape=(nn,), desc='current thrust', units='N')
        self.add_input(name='max_thrust', shape=(nn,), desc='maximum possible thrust', units='N')

        # Outputs
        self.add_output(name='tau', shape=(nn,), desc='throttle parameter', units=None)

        # Partials
        ar = np.arange(nn)
        self.declare_partials('tau', 'thrust', rows=ar, cols=ar)
        self.declare_partials('tau', 'max_thrust', rows=ar, cols=ar)

    def compute(self, inputs, outputs):
        outputs['tau'] = inputs['thrust'] / inputs['max_thrust']

    def compute_partials(self, inputs, partials):
        partials['tau', 'thrust'] = 1.0 / inputs['max_thrust']
        partials['tau', 'max_thrust'] = -inputs['thrust'] / inputs['max_thrust']**2



import os
import unittest

from openmdao.utils.testing_utils import use_tempdirs

import openmdao.api as om
import dymos as dm
from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE


@use_tempdirs
class TestBrachistochroneStaticGravity(unittest.TestCase):

    def _make_problem(self, tx):
        p = om.Problem(model=om.Group())
        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        #
        # Create a trajectory and add a phase to it
        #
        traj = p.model.add_subsystem('traj', dm.Trajectory())

        phase = traj.add_phase('phase0',
                               dm.Phase(ode_class=BrachistochroneODE,
                                        ode_init_kwargs={'static_gravity': True},
                                        transcription=tx))

        #
        # Set the variables
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', rate_source='xdot',
                        targets=None,
                        units='m',
                        fix_initial=True, fix_final=True, solve_segments=False)

        phase.add_state('y', rate_source='ydot',
                        targets=None,
                        units='m',
                        fix_initial=True, fix_final=True, solve_segments=False)

        phase.add_state('v', rate_source='vdot',
                        targets=['v'],
                        units='m/s',
                        fix_initial=True, fix_final=False, solve_segments=False)

        phase.add_control('theta', targets=['theta'],
                          continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_objective('time', loc='final', scaler=10)
        return p, phase

    def tearDown(self):
        for filename in ['total_coloring.pkl', 'SLSQP.out', 'SNOPT_print.out', 'SNOPT_summary.out']:
            if os.path.exists(filename):
                os.remove(filename)

    def test_conflicting_static_target(self):
        """ Tests that control cannot be connected to target tagged as 'dymos.static_target'. """
        p, phase = self._make_problem(dm.GaussLobatto(num_segments=10))

        phase.add_parameter('g', targets=['g'], static_target=False, opt=False)
        with self.assertRaises(ValueError) as e:
            p.setup()

        expected_msg = "Invalid parameter in phase `traj.phases.phase0`.\n" \
                       "Parameter `g` has invalid target(s).\n" \
                       "User has specified 'static_target = False' for parameter `g`,\nbut one or more " \
                       "targets is tagged with 'dymos.static_target':\n{'g'}"

        self.assertEqual(expected_msg, str(e.exception))

    def test_control_to_static_target_fails_gl(self):
        """ Tests that control cannot be connected to target tagged as 'dymos.static_target'. """
        p, phase = self._make_problem(dm.GaussLobatto(num_segments=10))

        phase.add_control('g', opt=False)
        with self.assertRaises(ValueError) as e:
            p.setup()

        expected_msg = "Control 'g' cannot be connected to its targets because one or more " \
                       "targets are tagged with 'dymos.static_target'."

        self.assertEqual(expected_msg, str(e.exception))

    def test_control_to_static_target_fails_radau(self):
        """ Tests that control cannot be connected to target tagged as 'dymos.static_target'. """
        p, phase = self._make_problem(dm.Radau(num_segments=10))

        phase.add_control('g', opt=False)
        with self.assertRaises(ValueError) as e:
            p.setup()

        expected_msg = "Control 'g' cannot be connected to its targets because one or more " \
                       "targets are tagged with 'dymos.static_target'."

        self.assertEqual(expected_msg, str(e.exception))

    def test_control_rate_to_static_target_fails_gl(self):
        """ Tests that control cannot be connected to target tagged as 'dymos.static_target'. """
        p, phase = self._make_problem(dm.GaussLobatto(num_segments=10))

        phase.add_control('foo', rate_targets=['g'], opt=False, shape=(1,), units='m/s')
        with self.assertRaises(ValueError) as e:
            p.setup()

        expected_msg = "Control rate of 'foo' cannot be connected to its targets because one or " \
                       "more targets are tagged with 'dymos.static_target'."

        self.assertEqual(expected_msg, str(e.exception))

    def test_control_rate_to_static_target_fails_radau(self):
        """ Tests that control cannot be connected to target tagged as 'dymos.static_target'. """
        p, phase = self._make_problem(dm.Radau(num_segments=10))

        phase.add_control('foo', rate_targets=['g'], opt=False, shape=(1,), units='m/s')
        with self.assertRaises(ValueError) as e:
            p.setup()

        expected_msg = "Control rate of 'foo' cannot be connected to its targets because one or " \
                       "more targets are tagged with 'dymos.static_target'."

        self.assertEqual(expected_msg, str(e.exception))

    def test_control_rate2_to_static_target_fails_gl(self):
        """ Tests that control cannot be connected to target tagged as 'dymos.static_target'. """
        p, phase = self._make_problem(dm.GaussLobatto(num_segments=10))

        phase.add_control('foo', rate2_targets=['g'], opt=False, shape=(1,), units='m/s')
        with self.assertRaises(ValueError) as e:
            p.setup()

        expected_msg = "Control rate2 of 'foo' cannot be connected to its targets because one or " \
                       "more targets are tagged with 'dymos.static_target'."

        self.assertEqual(expected_msg, str(e.exception))

    def test_control_rate2_to_static_target_fails_radau(self):
        """ Tests that control cannot be connected to target tagged as 'dymos.static_target'. """
        p, phase = self._make_problem(dm.Radau(num_segments=10))

        phase.add_control('foo', rate2_targets=['g'], opt=False, shape=(1,), units='m/s')
        with self.assertRaises(ValueError) as e:
            p.setup()

        expected_msg = "Control rate2 of 'foo' cannot be connected to its targets because one or " \
                       "more targets are tagged with 'dymos.static_target'."

        self.assertEqual(expected_msg, str(e.exception))

    def test_polynomial_control_to_static_target_fails_gl(self):
        """ Tests that control cannot be connected to target tagged as 'dymos.static_target'. """
        p, phase = self._make_problem(dm.GaussLobatto(num_segments=10))

        phase.add_control('g', opt=False, order=5, control_type='polynomial')
        with self.assertRaises(ValueError) as e:
            p.setup()

        expected_msg = "Control 'g' cannot be connected to its targets because one or " \
                       "more targets are tagged with 'dymos.static_target'."

        self.assertEqual(expected_msg, str(e.exception))

    def test_polynomial_control_to_static_target_fails_radau(self):
        """ Tests that control cannot be connected to target tagged as 'dymos.static_target'. """
        p, phase = self._make_problem(dm.Radau(num_segments=10))

        phase.add_control('g', opt=False, order=5, control_type='polynomial')
        with self.assertRaises(ValueError) as e:
            p.setup()

        expected_msg = "Control 'g' cannot be connected to its targets because one or " \
                       "more targets are tagged with 'dymos.static_target'."

        self.assertEqual(expected_msg, str(e.exception))

    def test_polynomial_control_rate_to_static_target_fails_gl(self):
        """ Tests that control cannot be connected to target tagged as 'dymos.static_target'. """
        p, phase = self._make_problem(dm.GaussLobatto(num_segments=10))

        phase.add_control('foo', opt=False, order=5, shape=(1,), units='m/s',
                          rate_targets=['g'], control_type='polynomial')
        with self.assertRaises(ValueError) as e:
            p.setup()

        expected_msg = "Control rate of 'foo' cannot be connected to its targets because one or " \
                       "more targets are tagged with 'dymos.static_target'."

        self.assertEqual(expected_msg, str(e.exception))

    def test_polynomial_control_rate_to_static_target_fails_radau(self):
        """ Tests that control cannot be connected to target tagged as 'dymos.static_target'. """
        p, phase = self._make_problem(dm.Radau(num_segments=10))

        phase.add_control('foo', opt=False, order=5, shape=(1,), units='m/s',
                          rate_targets=['g'], control_type='polynomial')
        with self.assertRaises(ValueError) as e:
            p.setup()

        expected_msg = "Control rate of 'foo' cannot be connected to its targets because one or " \
                       "more targets are tagged with 'dymos.static_target'."

        self.assertEqual(expected_msg, str(e.exception))

    def test_polynomial_control_rate2_to_static_target_fails_gl(self):
        """ Tests that control cannot be connected to target tagged as 'dymos.static_target'. """
        p, phase = self._make_problem(dm.GaussLobatto(num_segments=10))

        phase.add_control('foo', opt=False, order=5, shape=(1,), units='m/s',
                          rate2_targets=['g'], control_type='polynomial')
        with self.assertRaises(ValueError) as e:
            p.setup()

        expected_msg = "Control rate2 of 'foo' cannot be connected to its targets because one or " \
                       "more targets are tagged with 'dymos.static_target'."

        self.assertEqual(expected_msg, str(e.exception))

    def test_polynomial_control_rate2_to_static_target_fails_radau(self):
        """ Tests that control cannot be connected to target tagged as 'dymos.static_target'. """
        p, phase = self._make_problem(dm.Radau(num_segments=10))

        phase.add_control('foo', opt=False, order=5, shape=(1,), units='m/s',
                          rate2_targets=['g'], control_type='polynomial')
        with self.assertRaises(ValueError) as e:
            p.setup()

        expected_msg = "Control rate2 of 'foo' cannot be connected to its targets because one or " \
                       "more targets are tagged with 'dymos.static_target'."

        self.assertEqual(expected_msg, str(e.exception))


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest

from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse
from openmdao.utils.assert_utils import assert_near_equal

import numpy as np
import openmdao.api as om

import dymos as dm
from dymos.examples.cannonball.size_comp import CannonballSizeComp
from dymos.examples.cannonball.cannonball_ode import CannonballODE
from dymos.utils.misc import om_version


@use_tempdirs
class TestTwoPhaseCannonballLoadCase(unittest.TestCase):

    def _make_problem(self, ascent_only=False):
        """ Run ascent phase alone to create a restart file in which it is present. """
        p = om.Problem()

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = 'SLSQP'
        p.driver.options['print_results'] = False
        p.driver.declare_coloring()

        p.model.add_subsystem('size_comp', CannonballSizeComp(), promotes_inputs=['radius', 'dens'])
        p.model.set_input_defaults('dens', val=7.87, units='g/cm**3')
        p.model.add_design_var('radius', lower=0.01, upper=0.10, ref0=0.01, ref=0.10, units='m')

        traj = p.model.add_subsystem('traj', dm.Trajectory())

        transcription = dm.Radau(num_segments=5, order=3, compressed=True)
        ascent = dm.Phase(ode_class=CannonballODE, transcription=transcription)

        ascent = traj.add_phase('ascent', ascent)

        # All initial states except flight path angle are fixed
        # Final flight path angle is fixed (we will set it to zero so that the phase ends at apogee)
        ascent.set_time_options(fix_initial=True, duration_bounds=(1, 100), duration_ref=100, units='s')
        ascent.add_state('r', fix_initial=True, fix_final=False, units='m', rate_source='r_dot')
        ascent.add_state('h', fix_initial=True, fix_final=False,  units='m', rate_source='h_dot')
        ascent.add_state('gam', fix_initial=False, fix_final=True, units='rad', rate_source='gam_dot')
        ascent.add_state('v', fix_initial=False, fix_final=False, units='m/s', rate_source='v_dot')

        ascent.add_parameter('S', targets=['S'], units='m**2', static_target=True)
        ascent.add_parameter('mass', targets=['m'], units='kg', static_target=True)
        ascent.add_parameter('CD', targets=['CD'], units=None, static_target=True)

        # Limit the muzzle energy
        ascent.add_boundary_constraint('ke', loc='initial', units='J',
                                       upper=400000, lower=0, ref=100000, shape=(1,))

        # Temporary objective while building the problem
        ascent.add_objective('r', loc='final', scaler=-1.0)

        # Add internally-managed design parameters to the trajectory.
        traj.add_parameter('CD',
                           val=0.5, units=None, opt=False)

        # Add externally-provided design parameters to the trajectory.
        # In this case, we connect 'm' to pre-existing input parameters named 'mass' in each phase.
        if ascent_only:
            traj.add_parameter('m', units='kg', val=1.0,
                               targets={'ascent': 'mass'})
        else:
            traj.add_parameter('m', units='kg', val=1.0,
                               targets={'ascent': 'mass', 'descent': 'mass'})

        # In this case, by omitting targets, we're connecting these parameters to parameters
        # with the same name in each phase.
        traj.add_parameter('S', units='m**2', val=0.005)

        p.model.connect('size_comp.mass', 'traj.parameters:m')
        p.model.connect('size_comp.S', 'traj.parameters:S')

        if ascent_only:
            return p

        # Second Phase (descent)
        # transcription = dm.GaussLobatto(num_segments=5, order=3, compressed=True)
        descent = dm.Phase(ode_class=CannonballODE, transcription=transcription)
        traj.add_phase('descent', descent)

        # All initial states and time are free (they will be linked to the final states of ascent.
        # Final altitude is fixed (we will set it to zero so that the phase ends at ground impact)
        descent.set_time_options(initial_bounds=(.5, 100), duration_bounds=(.5, 100),
                                 duration_ref=100, units='s')
        descent.add_state('r', units='m', rate_source='r_dot')
        descent.add_state('h', fix_initial=False, fix_final=True,  units='m', rate_source='h_dot')
        descent.add_state('gam', fix_initial=False, fix_final=False, units='rad', rate_source='gam_dot')
        descent.add_state('v', fix_initial=False, fix_final=False, units='m/s', rate_source='v_dot')

        descent.add_parameter('S', targets=['S'], units='m**2', static_target=True)
        descent.add_parameter('mass', targets=['m'], units='kg', static_target=True)

        # Link Phases (link time and all state variables)
        # Note velocity is not included here.  Doing so is equivalent to linking kinetic energy,
        # and causes a duplicate row in the constraint jacobian.
        traj.link_phases(phases=['ascent', 'descent'], vars=['time', 'r', 'h', 'gam'], connected=False)

        traj.add_linkage_constraint('ascent', 'descent', 'ke', 'ke',
                                    ref=100000, connected=False)

        p.setup()

        return p

    @require_pyoptsparse(optimizer='SLSQP')
    def test_load_case_missing_phase(self):
        """ Test that loading a case with a missing phase gracefully skips that phase. """

        p_ascent_only = self._make_problem(ascent_only=True)

        p_ascent_only.setup()

        traj = p_ascent_only.model.traj
        ascent = traj.phases.ascent

        # Set pre-trajectory values
        p_ascent_only.set_val('radius', 0.05, units='m')
        p_ascent_only.set_val('dens', 7.87, units='g/cm**3')

        # Trajectory parameters
        traj.set_parameter_val('CD', 0.5)

        # Phase ascent times
        ascent.set_time_val(initial=0.0, duration=10.0)

        # Phase ascent state guesses
        ascent.set_state_val('r', [0, 100])
        ascent.set_state_val('h', [0, 100])
        ascent.set_state_val('v', [200, 150])
        ascent.set_state_val('gam', [25, 0], units='deg')

        dm.run_problem(p_ascent_only)

        p = self._make_problem()

        p.setup()

        ascent = p.model.traj.phases.ascent
        descent = p.model.traj.phases.descent

        # Set pre-trajectory values
        p.set_val('radius', 0.05, units='m')
        p.set_val('dens', 7.87, units='g/cm**3')

        # Trajectory parameters
        traj.set_parameter_val('CD', 0.5)

        ascent.set_time_val(initial=0.0, duration=10.0)

        # Phase ascent state guesses
        ascent.set_state_val('r', [0, 100])
        ascent.set_state_val('h', [0, 100])
        ascent.set_state_val('v', [200, 150])
        ascent.set_state_val('gam', [25, 0], units='deg')

        # Phase descent times
        descent.set_time_val(initial=10.0, duration=10.0)

        # Phase descent state guesses
        descent.set_state_val('r', [100, 200])
        descent.set_state_val('h', [100, 0])
        descent.set_state_val('v', [150, 200])
        descent.set_state_val('gam', [0, -45], units='deg')

        sol_db = 'dymos_solution.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = p_ascent_only.get_outputs_dir() / sol_db

        case = om.CaseReader(sol_db).get_case('final')

        if om_version()[0] < (3, 26, 1):
            dm.load_case(p, previous_solution=case)
        else:
            p.load_case(case)

        p.run_model()

        h_expected = np.array([0.00000000, 228.08533260, 434.90000516, 491.00415346,
                               592.60166806, 701.19757708, 730.03431228, 786.93071398,
                               848.54403903, 864.60323344, 895.67083544, 926.90913594,
                               934.25949414, 946.77982935, 954.88116543, 955.37021704])

        assert_near_equal(p.get_val('traj.ascent.states:h').ravel(), h_expected, tolerance=1.0E-5)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import openmdao.api as om

from ..grid_data import GridData
from .birkhoff_picard_iter_group import BirkhoffPicardIterGroup
from .multiple_shooting_update_comp import MultipleShootingUpdateComp


class MultipleShootingIterGroup(om.Group):
    """
    Class definition for the MultipleShootingIterGroup.

    This group allows for iteration of the initial or final state variable values in each segment
    (in forward shooting or backward shooting, respectively). This group feeds back new guesses
    for these "starting" state values based on the results of integration the causal segment

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def initialize(self):
        """
        Declare group options.
        """
        self.options.declare('state_options', types=dict,
                             desc='Dictionary of options for the states.')
        self.options.declare('time_units', types=str,
                             desc='Units for time in the phase.')
        self.options.declare('grid_data', types=GridData, desc='Container object for grid info.')
        self.options.declare('ode_class', default=None,
                             desc='Callable that instantiates the ODE system.',
                             recordable=False)
        self.options.declare('ode_init_kwargs', types=dict, default={},
                             desc='Keyword arguments provided when initializing the ODE System')
        self.options.declare('ode_nonlinear_solver', default=om.NonlinearBlockGS(maxiter=100, use_aitken=True, iprint=0),
                             desc='Nonlinear solver used to resolve Picard iteration.', recordable=False)
        self.options.declare('ode_linear_solver', default=om.DirectSolver(),
                             desc='Linear solver used to linearize the picard iteration subsystem.', recordable=False)
        self.options.declare('ms_nonlinear_solver', default=om.NonlinearBlockGS(maxiter=100, use_aitken=True, iprint=0),
                             desc='Nonlinear solver used to resolve Picard iteration.', recordable=False)
        self.options.declare('ms_linear_solver', default=om.DirectSolver(),
                             desc='Linear solver used to linearize the picard iteration subsystem.', recordable=False)
        self.options.declare('calc_exprs', types=dict, default={},
                             desc='phase calculation expressions.')
        self.options.declare('parameter_options', types=dict, default={},
                             desc='phase parameter options')

    def setup(self):
        """
        Define the structure of the control group.
        """
        gd = self.options['grid_data']
        state_options = self.options['state_options']
        time_units = self.options['time_units']
        ode_class = self.options['ode_class']
        ode_init_kwargs = self.options['ode_init_kwargs']
        calc_exprs = self.options['calc_exprs']
        parameter_options = self.options['parameter_options']

        seg_prop_group = self.add_subsystem('segment_prop_group',
                                            subsys=BirkhoffPicardIterGroup(grid_data=gd,
                                                                           state_options=state_options,
                                                                           time_units=time_units,
                                                                           ode_class=ode_class,
                                                                           ode_init_kwargs=ode_init_kwargs,
                                                                           calc_exprs=calc_exprs,
                                                                           parameter_options=parameter_options),
                                            promotes_inputs=['*'], promotes_outputs=['*'])

        seg_prop_group.nonlinear_solver = self.options['ode_nonlinear_solver']
        seg_prop_group.linear_solver = self.options['ode_linear_solver']

        self.nonlinear_solver = self.options['ms_nonlinear_solver']
        self.linear_solver = self.options['ms_linear_solver']

        self.add_subsystem('ms_update_comp',
                           MultipleShootingUpdateComp(grid_data=gd,
                                                      state_options=state_options),
                           promotes_inputs=['*'], promotes_outputs=['*'])

    def configure_io(self, phase):
        """
        I/O creation is delayed until configure so that we can determine shape and units for the states.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        state_options = self.options['state_options']

        segment_prop_group = self._get_subsystem('segment_prop_group')
        segment_prop_group.configure_io(phase)

        ms_update_comp = self._get_subsystem('ms_update_comp')
        ms_update_comp.configure_io(phase)

        for state_name, options in state_options.items():
            if options['solve_segments'] == 'forward':
                self.connect(f'seg_initial_states:{state_name}',
                             f'picard_update_comp.seg_initial_states:{state_name}')
            else:
                self.connect(f'seg_final_states:{state_name}',
                             f'picard_update_comp.seg_final_states:{state_name}')



from collections.abc import Sequence

import numpy as np

import openmdao.api as om

from .common import ControlInterpComp, ParameterComp
from ..utils.constants import INF_BOUND
from ..utils.indexing import get_constraint_flat_idxs
from ..utils.introspection import configure_states_introspection, get_promoted_vars, \
    configure_states_discovery, _configure_boundary_balance_introspection
from ..utils.misc import _unspecified, _format_phase_constraint_alias


class TranscriptionBase(object):
    """
    Base class for all dymos transcriptions.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def __init__(self, **kwargs):

        self._implicit_params = False
        self.grid_data = None

        self.options = om.OptionsDictionary()

        self.options.declare('num_segments', types=int, desc='Number of segments')
        self.options.declare('segment_ends', default=None, types=(Sequence, np.ndarray),
                             allow_none=True, desc='Locations of segment ends or None for equally '
                             'spaced segments')
        self.options.declare('order', default=3, types=(int, Sequence, np.ndarray),
                             desc='Order of the state transcription. The order of the control '
                                  'transcription is `order - 1`.')
        self.options.declare('compressed', default=True, types=bool,
                             desc='Use compressed transcription, meaning state and control values'
                                  'at segment boundaries are not duplicated on input.  This '
                                  'implicitly enforces value continuity between segments but in '
                                  'some cases may make the problem more difficult to solve.')

        self._declare_options()
        self.initialize()
        self.options.update(kwargs)
        self.init_grid()

        # Where to query var info.
        self._rhs_source = None

    def _declare_options(self):
        pass

    def initialize(self):
        """
        Declare transcription options.
        """
        pass

    def init_grid(self):
        """
        Setup the GridData object for the Transcription.
        """
        raise NotImplementedError(f'Transcription {self.__class__.__name__} does not implement method init_grid.')

    def setup_time(self, phase):
        """
        Setup up the time component and time extents for the phase.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        time_options = phase.time_options
        t_name = time_options['name']
        t_phase_name = f'{t_name}_phase'

        # Warn about invalid options
        phase.check_time_options()

        if phase.boundary_balance_options:
            self._implicit_params = True

        phase.add_subsystem('param_comp', subsys=ParameterComp(time_options=time_options),
                            promotes_inputs=['*'], promotes_outputs=['*'])

        for ts_name, ts_options in phase._timeseries.items():
            if t_name not in ts_options['outputs']:
                phase.add_timeseries_output(t_name, timeseries=ts_name)
            if t_phase_name not in ts_options['outputs'] and \
                    (phase.timeseries_options['include_t_phase'] or time_options['time_phase_targets']):
                phase.add_timeseries_output(t_phase_name, timeseries=ts_name)

    def configure_time(self, phase):
        """
        Configure the inputs/outputs on the time component.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        time_options = phase.time_options

        if not (time_options['input_initial'] or time_options['fix_initial']):
            lb, ub = time_options['initial_bounds']
            lb = -INF_BOUND if lb is None else lb
            ub = INF_BOUND if ub is None else ub

            phase.add_design_var('t_initial',
                                 lower=lb,
                                 upper=ub,
                                 scaler=time_options['initial_scaler'],
                                 adder=time_options['initial_adder'],
                                 ref0=time_options['initial_ref0'],
                                 ref=time_options['initial_ref'])

        if not (time_options['input_duration'] or time_options['fix_duration']):
            lb, ub = time_options['duration_bounds']
            lb = -INF_BOUND if lb is None else lb
            ub = INF_BOUND if ub is None else ub

            phase.add_design_var('t_duration',
                                 lower=lb,
                                 upper=ub,
                                 scaler=time_options['duration_scaler'],
                                 adder=time_options['duration_adder'],
                                 ref0=time_options['duration_ref0'],
                                 ref=time_options['duration_ref'])

    def setup_controls(self, phase):
        """
        Setup the control group.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        phase._check_control_options()

        if phase.control_options:
            control_comp = ControlInterpComp(control_options=phase.control_options,
                                             time_units=phase.time_options['units'],
                                             grid_data=self.grid_data)

            phase.add_subsystem('control_comp',
                                subsys=control_comp)

            phase.connect('t_duration_val', 'control_comp.t_duration')

    def configure_controls(self, phase):
        """
        Configure the inputs/outputs for the controls.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        control_prefix = 'controls:' if phase.timeseries_options['use_prefix'] else ''
        control_rate_prefix = 'control_rates:' if phase.timeseries_options['use_prefix'] else ''

        for name, options in phase.control_options.items():
            for ts_name, ts_options in phase._timeseries.items():
                if f'{control_prefix}{name}' not in ts_options['outputs']:
                    phase.add_timeseries_output(name, output_name=f'{control_prefix}{name}',
                                                timeseries=ts_name)
                if f'{control_rate_prefix}{name}_rate' not in ts_options['outputs'] and \
                        (phase.timeseries_options['include_control_rates'] or options['rate_targets']):
                    phase.add_timeseries_output(f'{name}_rate', output_name=f'{control_rate_prefix}{name}_rate',
                                                timeseries=ts_name)
                if f'{control_rate_prefix}{name}_rate2' not in ts_options['outputs'] and \
                        (phase.timeseries_options['include_control_rates'] or options['rate2_targets']):
                    phase.add_timeseries_output(f'{name}_rate2', output_name=f'{control_rate_prefix}{name}_rate2',
                                                timeseries=ts_name)

    def setup_parameters(self, phase):
        """
        Sets input defaults for parameters and optionally adds design variables.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        phase._check_parameter_options()
        param_prefix = 'parameters:' if phase.timeseries_options['use_prefix'] else ''
        include_params = phase.timeseries_options['include_parameters']

        for name, options in phase.parameter_options.items():
            if (options['include_timeseries'] is None and include_params) or options['include_timeseries']:
                for ts_name, ts_options in phase._timeseries.items():
                    if f'{param_prefix}{name}' not in ts_options['outputs']:
                        phase.add_timeseries_output(name, output_name=f'{param_prefix}{name}',
                                                    timeseries=ts_name)

    def configure_parameters(self, phase):
        """
        Configure parameter promotion.

        This method assumes that utils.introspection.configure_parameters_introspection has already populated
        the parameter options with the appropriate targets, units, shape, and static_target fields.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        if phase.parameter_options:
            param_comp = phase._get_subsystem('param_comp')

            for name, options in phase.parameter_options.items():
                param_comp.add_parameter(name, val=options['val'], shape=options['shape'], units=options['units'])
                if options['opt']:
                    lb = -INF_BOUND if options['lower'] is None else options['lower']
                    ub = INF_BOUND if options['upper'] is None else options['upper']
                    phase.add_design_var(name=f'parameters:{name}',
                                         lower=lb,
                                         upper=ub,
                                         scaler=options['scaler'],
                                         adder=options['adder'],
                                         ref0=options['ref0'],
                                         ref=options['ref'])

                for tgts, src_idxs in self.get_parameter_connections(name, phase):
                    phase.connect(f'parameter_vals:{name}', tgts, src_indices=src_idxs,
                                  flat_src_indices=True)

    def setup_states(self, phase):
        """
        Setup the states for this transcription.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        raise NotImplementedError(f'Transcription {self.__class__.__name__} does not implement method setup_states.')

    def configure_states_introspection(self, phase):
        """
        Perform introspection on the RHS system for the states for this transcription.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        ode = self._get_ode(phase)
        try:
            configure_states_introspection(phase.state_options, phase.time_options, phase.control_options,
                                           phase.parameter_options, ode)
        except (ValueError, RuntimeError) as e:
            raise RuntimeError(f'Error during configure_states_introspection in phase {phase.pathname}.') from e

    def configure_states_discovery(self, phase):
        """
        Perform introspection on the RHS system to automatically discover states for this transcription.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        ode = self._get_ode(phase)
        try:
            configure_states_discovery(phase.state_options, ode)
        except (ValueError, RuntimeError) as e:
            raise RuntimeError(f'Error during configure_states_discovery in phase {phase.pathname}.') from e

    def configure_states(self, phase):
        """
        Configure the states for this transcription.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        state_prefix = 'states:' if phase.timeseries_options['use_prefix'] else ''
        state_rate_prefix = 'state_rates:' if phase.timeseries_options['use_prefix'] else ''

        for name, options in phase.state_options.items():
            for ts_name, ts_options in phase._timeseries.items():
                if f'{state_prefix}{name}' not in ts_options['outputs']:
                    phase.add_timeseries_output(name, output_name=f'{state_prefix}{name}',
                                                timeseries=ts_name, tags='state')
                if options['rate_source'] and phase.timeseries_options['include_state_rates']:
                    output_name = f'{state_rate_prefix}{name}' if state_rate_prefix else options['rate_source']
                    if output_name not in ts_options['outputs']:
                        phase.add_timeseries_output(name=options['rate_source'],
                                                    output_name=output_name,
                                                    timeseries=ts_name, tags='state_rate')

    def setup_ode(self, phase):
        """
        Setup the ode for this transcription.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        raise NotImplementedError(f'Transcription {self.__class__.__name__} does not implement method setup_ode.')

    def setup_boundary_balance(self, phase):
        """
        Setup the implicit computation of the phase boundary balance.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        if phase.boundary_balance_options:
            boundary_balance_comp = om.BalanceComp()
            phase.add_subsystem('boundary_balance_comp', boundary_balance_comp, promotes_outputs=['*'])

    def configure_boundary_balance(self, phase):
        """
        Configure the implicit computation of the phase boundary balance.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        param_balance_comp = phase._get_subsystem('boundary_balance_comp')

        _configure_boundary_balance_introspection(phase)

        for param, options in phase.boundary_balance_options.items():
            name = options['name']
            tgt_val = options['tgt_val']
            loc = options['loc']
            index = [options['index']] if np.isscalar(options['index']) else options['index']

            # Get the indices to connect based on loc.
            if loc == 'final':
                src_idxs = om.slicer[(-1, *index)]
            elif loc == 'initial':
                src_idxs = om.slicer[(0, *index)]
            else:
                raise ValueError(f'{phase.msginfo}: Value of `loc` for boundary balance `{param}` '
                                 'must be one of `initial` or `final`, but got `{loc}` instead.')

            # Create the arguments for the balance comp.
            bal_kwargs = {key: options for key, options in options.items()}
            try:
                output_name = bal_kwargs.pop('output_name')
            except KeyError:
                output_name = name.split('.')[-1]
            bal_kwargs.pop('param')
            bal_kwargs.pop('name')
            bal_kwargs.pop('tgt_val')
            bal_kwargs.pop('loc')
            bal_kwargs.pop('index')
            bal_kwargs['rhs_val'] = tgt_val
            bal_kwargs['lhs_name'] = output_name

            prom_param_name = f'parameters:{param}' if param in phase.parameter_options else param

            # Now configure the balance.
            param_balance_comp.add_balance(name=prom_param_name, **bal_kwargs)

            var_type = phase.classify_var(name)
            if var_type == 'ode':
                if name not in phase._timeseries['timeseries']['outputs']:
                    phase.add_timeseries_output(name, output_name=output_name, units=bal_kwargs.get('eq_units', _unspecified))
            if var_type == 'state' and name.startswith('initial_states:') or name.startswith('final_states:'):
                phase.promotes('boundary_balance_comp', inputs=[output_name])
            else:
                phase.connect(f'timeseries.{output_name}', f'boundary_balance_comp.{output_name}',
                              src_indices=src_idxs)

    def setup_solvers(self, phase):
        """
        Setup the solvers for this transcription.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        raise NotImplementedError(f'Transcription {self.__class__.__name__} does not implement method setup_solvers.')

    def configure_solvers(self, phase, requires_solvers=None):
        """
        Configure the solvers.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        requires_solvers : dict[str: bool]
            A dictionary mapping a string descriptor of a reason why a solver is required,
            and whether a solver is required.
        """
        if not phase.options['auto_solvers']:
            return

        req_solvers = {'implicit parameters': self._implicit_params}

        if requires_solvers is not None:
            req_solvers.update(requires_solvers)

        reasons_txt = '\n'.join(f'    {key}: {val}' for key, val in req_solvers.items())
        warn = False

        if any(req_solvers.values()):
            msg = (f'{phase.msginfo}\n'
                   f'  Non-default solvers are required\n'
                   f'{reasons_txt}\n')
            if isinstance(phase.nonlinear_solver, om.NonlinearRunOnce):
                msg += (f'  Setting `{phase.pathname}.nonlinear_solver = om.NewtonSolver(iprint=0, '
                        f'solve_subsystems=True, maxiter=1000, stall_limit=3)`\n'
                        f'  Explicitly set {phase.pathname}.nonlinear_solver to override.\n')
                warn = True
                phase.nonlinear_solver = om.NewtonSolver(iprint=0)
                phase.nonlinear_solver.options['solve_subsystems'] = True
                phase.nonlinear_solver.options['maxiter'] = 100
                phase.nonlinear_solver.options['stall_limit'] = 3
                phase.nonlinear_solver.linesearch = om.ArmijoGoldsteinLS()

            if isinstance(phase.linear_solver, om.LinearRunOnce):
                warn = True
                msg += (f'  Setting `{phase.pathname}.linear_solver = om.DirectSolver(iprint=2)`\n'
                        f'  Explicitly set {phase.pathname}.linear_solver to override.\n')
                phase.linear_solver = om.DirectSolver(iprint=0)

            if warn:
                msg += f'  Set `{phase.pathname}.options["auto_solvers"] = False` to disable this behavior.'
                om.issue_warning(msg)

    def setup_timeseries_outputs(self, phase):
        """
        Setup the timeseries for this transcription.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        raise NotImplementedError(f'Transcription {self.__class__.__name__} does not implement method '
                                  f'setup_timeseries_outputs.')

    def configure_timeseries_outputs(self, phase):
        """
        Create connections from time series to all post-introspection sources.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        for timeseries_name, timeseries_options in phase._timeseries.items():
            timeseries_comp = phase._get_subsystem(timeseries_name)

            for input_name, src, src_idxs in timeseries_comp._configure_io(timeseries_options):
                phase.connect(src_name=src,
                              tgt_name=f'{timeseries_name}.{input_name}',
                              src_indices=src_idxs)

    def configure_boundary_constraints(self, phase):
        """
        Configures the boundary constraints.

        Adds BoundaryConstraintComp for initial and/or final boundary constraints if necessary
        and issues appropriate connections.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """

        for ibc in phase._initial_boundary_constraints:
            con_output, constraint_kwargs = self._get_constraint_kwargs('initial', ibc, phase)
            phase.add_constraint(con_output, **constraint_kwargs)

        for fbc in phase._final_boundary_constraints:
            con_output, constraint_kwargs = self._get_constraint_kwargs('final', fbc, phase)
            phase.add_constraint(con_output, **constraint_kwargs)

    def _get_constraint_kwargs(self, constraint_type, options, phase):
        """
        Given the constraint options provide the keyword arguments for the OpenMDAO add_constraint method.

        Parameters
        ----------
        constraint_type : str
            One of 'initial', 'final', or 'path'.
        options : dict
            The constraint options.
        phase : Phase
            The dymos phase to which the constraint applies.

        Returns
        -------
        con_output : str
            The phase-relative path being constrained.
        constraint_kwargs : dict
            Keyword arguments for the OpenMDAO add_constraint method.
        """
        num_nodes = self._get_num_timeseries_nodes()

        constraint_kwargs = {key: options for key, options in options.items()}

        # Determine the path to the variable which we will be constraining
        var = options['name']
        var_type = phase.classify_var(var)

        # These are the flat indices at a single point in time used
        # in either initial, final, or path constraints.
        idxs_in_initial = phase._indices_in_constraints(var, 'initial')
        idxs_in_final = phase._indices_in_constraints(var, 'final')
        idxs_in_path = phase._indices_in_constraints(var, 'path')

        size = np.prod(options['shape'], dtype=int)

        flat_idxs = get_constraint_flat_idxs(options)

        # Now we need to convert the indices given by the user at any given point
        # to flat indices to be given to OpenMDAO as flat indices spanning the phase.
        if var_type == 'parameter':
            if any([idxs_in_initial.intersection(idxs_in_final),
                    idxs_in_initial.intersection(idxs_in_path),
                    idxs_in_final.intersection(idxs_in_path)]):
                raise RuntimeError(f'In phase {phase.pathname}, parameter `{var}` is subject to multiple boundary '
                                   f'or path constraints.\nParameters are single values that do not change in '
                                   f'time, and may only be used in a single boundary or path constraint.')
            constraint_kwargs['indices'] = flat_idxs
        else:
            if constraint_type == 'initial':
                constraint_kwargs['indices'] = flat_idxs
            elif constraint_type == 'final':
                constraint_kwargs['indices'] = size * (num_nodes - 1) + flat_idxs
            else:
                # This is a path constraint.
                # Remove any flat indices involved in an initial constraint from the path constraint
                flat_idxs_set = set(flat_idxs)
                idxs_not_in_initial = list(flat_idxs_set - idxs_in_initial)

                # Remove any flat indices involved in the final constraint from the path constraint
                idxs_not_in_final = list(flat_idxs_set - idxs_in_final)
                idxs_not_in_final = (size * (num_nodes - 1) + np.asarray(idxs_not_in_final)).tolist()

                intermediate_idxs = []
                for i in range(1, num_nodes - 1):
                    intermediate_idxs.extend(size * i + flat_idxs)

                constraint_kwargs['indices'] = idxs_not_in_initial + intermediate_idxs + idxs_not_in_final

        con_path = constraint_kwargs.pop('constraint_path')
        con_name = constraint_kwargs.pop('constraint_name')

        constraint_kwargs['alias'] = _format_phase_constraint_alias(phase, con_name,
                                                                    constraint_type,
                                                                    options['indices'])
        constraint_kwargs.pop('name')
        constraint_kwargs.pop('shape')
        constraint_kwargs['flat_indices'] = True

        return con_path, constraint_kwargs

    def configure_path_constraints(self, phase):
        """
        Handle the common operations for configuration of the path constraints.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        for pc in phase._path_constraints:
            con_output, constraint_kwargs = self._get_constraint_kwargs('path', pc, phase)
            phase.add_constraint(con_output, **constraint_kwargs)

    def configure_objective(self, phase):
        """
        Find the path of the objective(s) and add them.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        for name, options in phase._objectives.items():
            index = options['index']
            loc = options['loc']

            obj_path, shape, _, _ = self._get_response_src(name, loc, phase)

            shape = options['shape'] if shape is None else shape

            size = int(np.prod(shape))

            if size > 1 and index is None:
                raise ValueError(f'Objective variable is non-scaler {shape} but no index specified for objective')

            idx = 0 if index is None else index
            if idx < 0:
                idx = size + idx

            if idx >= size or idx < -size:
                raise ValueError('Objective index={0}, but the shape of the objective '
                                 'variable is {1}'.format(index, shape))

            if loc == 'final':
                obj_index = -size + idx
            elif loc == 'initial':
                obj_index = idx
            else:
                raise ValueError('Invalid value for objective loc: {0}. Must be '
                                 'one of \'initial\' or \'final\'.'.format(loc))

            from ..phase import Phase
            super(Phase, phase).add_objective(obj_path, ref=options['ref'], ref0=options['ref0'],
                                              index=obj_index, flat_indices=True, adder=options['adder'],
                                              scaler=options['scaler'],
                                              parallel_deriv_color=options['parallel_deriv_color'])

    def _get_response_src(self, name, loc, phase, ode_outputs=None):
        """
        Return the path to the variable that will be used as a response..

        Parameters
        ----------
        var : str
            Name of the variable to be used as the response.
        loc : str
            The location of the response in the phase ['initial', 'final'].
        phase : dymos.Phase
            Phase object containing in which the objective resides.
        ode_outputs : dict or None
            A dictionary of ODE outputs as returned by get_promoted_vars.

        Returns
        -------
        obj_path : str
            Path to the source.
        shape : tuple
            Source shape.
        units : str
            Source units.
        linear : bool
            True if the objective quantity is linear.
        """
        raise NotImplementedError(f'Transcription {self.__class__.__name__} does not implement method '
                                  '_get_response_src.')

    def _get_rate_source_path(self, name, loc, phase):
        raise NotImplementedError(f'Transcription {self.__class__.__name__} does not implement method '
                                  '_get_rate_source_path.')

    def _get_ode(self, phase):
        """
        Returns an instance of the ODE used in the phase that can be interrogated for IO metadata.

        Parameters
        ----------
        phase : dm.Phase
            The Phase instance to which this transcription applies

        Returns
        -------
        ode : om.System
            The OpenMDAO system which serves as the ODE for the given Phase.

        """
        ode = phase._get_subsystem(self._rhs_source)
        if ode is None:
            raise AttributeError(f'Phase ODE subsystem {self._rhs_source} not found.')
        return phase._get_subsystem(self._rhs_source)

    def get_parameter_connections(self, name, phase):
        """
        Returns info about a parameter's target connections in the phase.

        Parameters
        ----------
        name : str
            The name of the parameter for which connection information is desired.
        phase : dymos.Phase
            The phase object to which this transcription applies.

        Returns
        -------
        list of (paths, indices)
            A list containing a tuple of target paths and corresponding src_indices to which the
            given design variable is to be connected.
        """
        raise NotImplementedError(f'Transcription {self.__class__.__name__} does not implement method'
                                  f'get_parameter_connections.')

    def is_static_ode_output(self, var, phase, num_nodes):
        """
        Test whether the given output is a static output of the ODE.

        A variable is considered static if it's first dimension is different than the
        number of nodes in the ODE.

        Parameters
        ----------
        var : str
            The ode-relative path of the variable of interest.
        phase : dymos.Phase or dict
            The phase to which this transcription applies or a dict of the ODE outputs as returned by get_promoted_vars.
        num_nodes : int
            The number of nodes in the ODE.

        Returns
        -------
        bool
            True if the given variable is a static output, otherwise False if it is dynamic.

        Raises
        ------
        KeyError
            KeyError is raised if the given variable isn't present in the ode outputs.
        """
        if isinstance(phase, dict):
            ode_outputs = phase
        else:
            ode_outputs = get_promoted_vars(self._get_ode(phase), 'output')
        ode_shape = ode_outputs[var]['shape']
        return ode_shape[0] != num_nodes

    def _requires_continuity_constraints(self, phase):
        """
        Tests whether state and/or control and/or control rate continuity are required.

        Parameters
        ----------
        phase : dymos.Phase
            The phase to which this transcription applies.

        Returns
        -------
        state_continuity : bool
            True if any state continuity is required to be enforced.
        control_continuity : bool
            True if any control value continuity is required to be enforced.
        control_rate_continuity : bool
            True if any control rate continuity is required to be enforced.
        """
        raise NotImplementedError(f'The transcription {self.__class__} does not provide an '
                                  f'implementation of _requires_continuity_constraints')

    def _phase_set_state_val(self, phase, name, vals, times, interpolation_kind):
        """
        Method to interpolate the provided input and return the variables that need to be set
        along with their appropriate value.

        Parameters
        ----------
        phase : dymos.Phase
            The phase to which this transcription applies.
        name : str
            The name of the phase variable to be set.
        vals : ndarray or Sequence or float
            Array of control/state/parameter values.
        times : ndarray or Sequence or None
            Array of integration variable values.
        interpolation_kind : str
            Specifies the kind of interpolation, as per the scipy.interpolate package.
            One of ('linear', 'nearest', 'zero', 'slinear', 'quadratic', 'cubic'
            where 'zero', 'slinear', 'quadratic' and 'cubic' refer to a spline
            interpolation of zeroth, first, second or third order) or as an
            integer specifying the order of the spline interpolator to use.
            Default is 'linear'.

        Returns
        -------
        input_data : dict
            Dict containing the values that need to be set in the phase

        """
        raise NotImplementedError(f'Transcription {self.__class__.__name__} does not implement method '
                                  '_phase_set_val.')

    def _get_num_timeseries_nodes(self):
        """
        Returns the number of nodes in the default timeseries for this transcription.

        Returns
        -------
        int
            The number of nodes in the default timeseries for this transcription.
        """
        raise NotImplementedError(f'Transcription {self.__class__.__name__} does not implement method '
                                  '_get_num_timeseries_nodes.')



import unittest

import numpy as np
from numpy.testing import assert_almost_equal

from dymos.utils.lgl import lgl

# Known solutions
x_i = {2: [-1.0, 1.0],
       3: [-1.0, 0.0, 1.0],
       4: [-1.0, -np.sqrt(5)/5, np.sqrt(5)/5, 1.0],
       5: [-1.0, -np.sqrt(21)/7, 0.0, np.sqrt(21)/7, 1.0],
       6: [-1.0, -np.sqrt((7 + 2 * np.sqrt(7)) / 21), -np.sqrt((7 - 2 * np.sqrt(7)) / 21),
           np.sqrt((7 - 2 * np.sqrt(7)) / 21), np.sqrt((7 + 2 * np.sqrt(7)) / 21), 1.0]}

w_i = {2: [1.0, 1.0],
       3: [1/3, 4/3, 1/3],
       4: [1/6, 5/6, 5/6, 1/6],
       5: [1/10, 49/90, 32/45, 49/90, 1/10],
       6: [1/15, (14 - np.sqrt(7))/30, (14 + np.sqrt(7))/30,
           (14 + np.sqrt(7))/30, (14 - np.sqrt(7))/30, 1/15]}


class TestLGL(unittest.TestCase):

    def test_nodes_and_weights_2(self):
        x_2, w_2 = lgl(2)
        assert_almost_equal(x_2, x_i[2], decimal=6)
        assert_almost_equal(w_2, w_i[2], decimal=6)

    def test_nodes_and_weights_3(self):
        x_3, w_3 = lgl(3)
        assert_almost_equal(x_3, x_i[3], decimal=6)
        assert_almost_equal(w_3, w_i[3], decimal=6)

    def test_nodes_and_weights_4(self):
        x_4, w_4 = lgl(4)
        assert_almost_equal(x_4, x_i[4], decimal=6)
        assert_almost_equal(w_4, w_i[4], decimal=6)

    def test_nodes_and_weights_5(self):
        x_5, w_5 = lgl(5)
        assert_almost_equal(x_5, x_i[5], decimal=6)
        assert_almost_equal(w_5, w_i[5], decimal=6)

    def test_nodes_and_weights_6(self):
        x_6, w_6 = lgl(6)
        assert_almost_equal(x_6, x_i[6], decimal=6)
        assert_almost_equal(w_6, w_i[6], decimal=6)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import openmdao.api as om

from .birkhoff_picard_update_comp import PicardUpdateComp
from .states_comp import StatesComp

from ..grid_data import GridData
from ...utils.ode_utils import _make_ode_system


class BirkhoffPicardIterGroup(om.Group):
    """
    Class definition for the BirkhoffIterGroup.

    This group allows for iteration of the state variables and initial _or_ final value of the state
    depending on the direction of the solve.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._implicit_outputs = set()

    def initialize(self):
        """
        Declare group options.
        """
        self.options.declare('state_options', types=dict,
                             desc='Dictionary of options for the states.')
        self.options.declare('time_units', types=str,
                             desc='Units for time in the phase.')
        self.options.declare('grid_data', types=GridData, desc='Container object for grid info.')
        self.options.declare('ode_class', default=None,
                             desc='Callable that instantiates the ODE system.',
                             recordable=False)
        self.options.declare('ode_init_kwargs', types=dict, default={},
                             desc='Keyword arguments provided when initializing the ODE System')
        self.options.declare('calc_exprs', types=dict, default={},
                             desc='phase calculation expressions.')
        self.options.declare('parameter_options', types=dict, default={},
                             desc='phase parameter options')

    def setup(self):
        """
        Define the structure of the control group.
        """
        gd = self.options['grid_data']
        nn = gd.subset_num_nodes['all']
        state_options = self.options['state_options']
        time_units = self.options['time_units']

        self.add_subsystem('states_comp',
                           subsys=StatesComp(grid_data=gd, state_options=state_options))

        ode = _make_ode_system(ode_class=self.options['ode_class'],
                               num_nodes=nn,
                               calc_exprs=self.options['calc_exprs'],
                               ode_init_kwargs=self.options['ode_init_kwargs'],
                               parameter_options=self.options['parameter_options'])

        self.add_subsystem('ode_all', subsys=ode)

        self.add_subsystem('picard_update_comp',
                           subsys=PicardUpdateComp(grid_data=gd,
                                                   state_options=state_options,
                                                   time_units=time_units))

    def configure_io(self, phase):
        """
        I/O creation is delayed until configure so that we can determine shape and units for the states.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        states_comp = self._get_subsystem('states_comp')
        states_comp.configure_io(phase)

        self.promotes('states_comp', inputs=['*'], outputs=['*'])

        picard_update_comp = self._get_subsystem('picard_update_comp')
        picard_update_comp.configure_io(phase)

        state_options = self.options['state_options']

        for name, options in state_options.items():
            rate_source = options['rate_source']

            for tgt in options['targets']:
                self.connect(f'state_val:{name}', f'ode_all.{tgt}')

            try:
                rate_source_var = options['rate_source']
            except RuntimeError:
                raise ValueError(f"state '{name}' in phase '{phase.name}' was not given a "
                                 "rate_source")

            # Note the rate source must be shape-compatible with the state
            var_type = phase.classify_var(rate_source_var)

            if var_type == 'ode':
                self.connect(f'ode_all.{rate_source}', f'picard_update_comp.f_computed:{name}')

            promotes = [f'states:{name}']
            if options['solve_segments'] == 'forward':
                promotes += [f'final_states:{name}']
            else:
                promotes += [f'initial_states:{name}']

            self.promotes('picard_update_comp', any=promotes)
